import pytest
import codecs
import decimal

from faster_eth_utils import (
    decode_hex,
    int_to_big_endian,
    is_address,
    is_boolean,
    is_bytes,
    is_integer,
    is_number,
    is_text,
    to_canonical_address,
    to_checksum_address,
    to_normalized_address,
)
from hypothesis import (
    example,
    given,
    settings,
    strategies as st,
)

from faster_eth_abi.encoding import (
    AddressEncoder,
    BooleanEncoder,
    BytesEncoder,
    ByteStringEncoder,
    SignedFixedEncoder,
    SignedIntegerEncoder,
    TextStringEncoder,
    TupleEncoder,
    UnsignedFixedEncoder,
    UnsignedIntegerEncoder,
    encode_uint_256,
)
from faster_eth_abi.exceptions import (
    EncodingTypeError,
    IllegalValue,
    ValueOutOfBounds,
)
from faster_eth_abi.utils.numeric import (
    TEN,
    abi_decimal_context,
    ceil32,
    compute_signed_fixed_bounds,
    compute_signed_integer_bounds,
    compute_unsigned_fixed_bounds,
    compute_unsigned_integer_bounds,
)
from faster_eth_abi.utils.padding import (
    fpad,
    zpad,
    zpad_right,
)


@given(
    bool_value=st.one_of(st.booleans(), st.none()),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_encode_boolean(bool_value, data_byte_size):
    encoder = BooleanEncoder(
        data_byte_size=data_byte_size,
    )

    if not is_boolean(bool_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(bool_value)
        assert "BooleanEncoder" in str(exception_info.value)
        return

    expected_value = zpad(b"\x01" if bool_value else b"\x00", data_byte_size)
    encoded_value = encoder(bool_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@example(integer_value=-1, value_bit_size=8, data_byte_size=1)
@given(
    integer_value=st.one_of(st.integers(), st.none()),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_encode_unsigned_integer(integer_value, value_bit_size, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError) as exception_info:
            UnsignedIntegerEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = UnsignedIntegerEncoder(
        value_bit_size=value_bit_size,
        data_byte_size=data_byte_size,
    )
    lower_bound, upper_bound = compute_unsigned_integer_bounds(value_bit_size)

    if not is_integer(integer_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(integer_value)
        assert "UnsignedInteger" in str(exception_info.value)
        return
    elif integer_value < lower_bound or integer_value > upper_bound:
        with pytest.raises(ValueOutOfBounds):
            encoder(integer_value)
        return

    if integer_value >= 0:
        expected_value = zpad(int_to_big_endian(integer_value), data_byte_size)
    else:
        expected_value = fpad(int_to_big_endian(integer_value), data_byte_size)

    encoded_value = encoder(integer_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    integer_value=st.one_of(st.integers(), st.none()),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_encode_signed_integer(integer_value, value_bit_size, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            SignedIntegerEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = SignedIntegerEncoder(
        value_bit_size=value_bit_size,
        data_byte_size=data_byte_size,
    )

    lower_bound, upper_bound = compute_signed_integer_bounds(value_bit_size)

    if not is_integer(integer_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(integer_value)
        assert "SignedInteger" in str(exception_info.value)
        return
    elif integer_value < lower_bound or integer_value > upper_bound:
        with pytest.raises(ValueOutOfBounds):
            encoder(integer_value)
        return

    if integer_value >= 0:
        expected_value = zpad(
            int_to_big_endian(integer_value % 2**value_bit_size), data_byte_size
        )
    else:
        expected_value = fpad(
            int_to_big_endian(integer_value % 2**value_bit_size), data_byte_size
        )

    encoded_value = encoder(integer_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    address_value=st.one_of(
        st.none(),
        st.binary(min_size=20, max_size=20),
        st.binary(min_size=20, max_size=20).map(to_normalized_address),
        st.binary(min_size=20, max_size=20).map(to_checksum_address),
    ),
    value_bit_size=st.integers(min_value=19, max_value=21).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=18, max_value=32),
)
def test_encode_address(address_value, value_bit_size, data_byte_size):
    if value_bit_size != 20 * 8:
        with pytest.raises(ValueError):
            AddressEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            AddressEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = AddressEncoder(
        value_bit_size=value_bit_size,
        data_byte_size=data_byte_size,
    )

    if not is_address(address_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(address_value)
        assert "AddressEncoder" in str(exception_info.value)
        return

    expected_value = zpad(to_canonical_address(address_value), data_byte_size)
    encoded_value = encoder(address_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    bytes_value=st.one_of(
        st.none(),
        st.binary(min_size=0, max_size=40),
    ),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    data_byte_size=st.integers(min_value=18, max_value=32),
)
def test_encode_bytes_xx(bytes_value, value_bit_size, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            AddressEncoder(
                value_bit_size=value_bit_size,
                data_byte_size=data_byte_size,
            )
        return

    encoder = BytesEncoder(
        value_bit_size=value_bit_size,
        data_byte_size=data_byte_size,
    )

    if not is_bytes(bytes_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(bytes_value)
        assert "BytesEncoder" in str(exception_info.value)
        return
    elif len(bytes_value) * 8 > value_bit_size:
        with pytest.raises(ValueOutOfBounds):
            encoder(bytes_value)
        return

    expected_value = zpad_right(bytes_value, data_byte_size)
    encoded_value = encoder(bytes_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    string_value=st.one_of(
        st.none(),
        st.text(min_size=0, max_size=256),
        st.binary(min_size=0, max_size=256),
    ),
)
def test_encode_byte_string(string_value):
    encoder = ByteStringEncoder()

    if not is_bytes(string_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(string_value)
        assert "ByteStringEncoder" in str(exception_info.value)
        return

    expected_value = encode_uint_256(len(string_value)) + (
        zpad_right(string_value, ceil32(len(string_value))) if string_value else b""
    )
    encoded_value = encoder(string_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    string_value=st.one_of(
        st.none(),
        st.text(min_size=0, max_size=256),
        st.binary(min_size=0, max_size=256),
    ),
)
def test_encode_text_string(string_value):
    encoder = TextStringEncoder()

    if not is_text(string_value):
        with pytest.raises(EncodingTypeError) as exception_info:
            encoder(string_value)
        assert "TextStringEncoder" in str(exception_info.value)
        return

    string_value_as_bytes = codecs.encode(string_value, "utf8")

    expected_value = (
        encode_uint_256(len(string_value_as_bytes))
        + (
            zpad_right(
                string_value_as_bytes,
                ceil32(len(string_value_as_bytes)),
            )
            if string_value
            else b""
        )
        if string_value
        else b"\x00" * 32
    )
    encoded_value = encoder(string_value)

    assert encoded_value == expected_value


@settings(max_examples=250)
@given(
    value=st.one_of(st.integers(), st.decimals(), st.none()),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    frac_places=st.integers(min_value=1, max_value=80),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
@example(
    value=decimal.Decimal("5.33"), value_bit_size=8, frac_places=1, data_byte_size=1
)
def test_encode_unsigned_fixed(value, value_bit_size, frac_places, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        pattern = r"Value byte size exceeds data size"
        with pytest.raises(ValueError, match=pattern):
            UnsignedFixedEncoder(
                value_bit_size=value_bit_size,
                frac_places=frac_places,
                data_byte_size=data_byte_size,
            )
        return

    encoder = UnsignedFixedEncoder(
        value_bit_size=value_bit_size,
        frac_places=frac_places,
        data_byte_size=data_byte_size,
    )

    if not is_number(value):
        pattern = (
            r"Value `None` of type .*NoneType.* cannot be "
            r"encoded by UnsignedFixedEncoder"
        )
        with pytest.raises(EncodingTypeError, match=pattern):
            encoder(value)
        return

    if UnsignedFixedEncoder.illegal_value_fn(value):
        pattern = (
            r"Value .*(NaN|Infinity|-Infinity).* cannot be "
            r"encoded by UnsignedFixedEncoder"
        )
        with pytest.raises(IllegalValue, match=pattern):
            encoder(value)
        return

    lower, upper = compute_unsigned_fixed_bounds(value_bit_size, frac_places)
    if value < lower or value > upper:
        pattern = (
            r"Value .* cannot be encoded by UnsignedFixedEncoder: "
            r"Cannot be encoded in .* bits"
        )
        with pytest.raises(ValueOutOfBounds, match=pattern):
            encoder(value)
        return

    with decimal.localcontext(abi_decimal_context):
        residue = value % (TEN**-frac_places)
    if residue > 0:
        pattern = (
            r"Value .* cannot be encoded by UnsignedFixedEncoder: "
            r"residue .* outside allowed"
        )
        with pytest.raises(IllegalValue, match=pattern):
            encoder(value)
        return

    # Ensure no exception
    encoder(value)


@settings(max_examples=250)
@given(
    value=st.one_of(st.integers(), st.decimals(), st.none()),
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    frac_places=st.integers(min_value=1, max_value=80),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_encode_signed_fixed(value, value_bit_size, frac_places, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        pattern = r"Value byte size exceeds data size"
        with pytest.raises(ValueError, match=pattern):
            SignedFixedEncoder(
                value_bit_size=value_bit_size,
                frac_places=frac_places,
                data_byte_size=data_byte_size,
            )
        return

    encoder = SignedFixedEncoder(
        value_bit_size=value_bit_size,
        frac_places=frac_places,
        data_byte_size=data_byte_size,
    )

    if not is_number(value):
        pattern = (
            r"Value `None` of type .*NoneType.* cannot be encoded by SignedFixedEncoder"
        )
        with pytest.raises(EncodingTypeError, match=pattern):
            encoder(value)
        return

    if SignedFixedEncoder.illegal_value_fn(value):
        pattern = (
            r"Value .*(NaN|Infinity|-Infinity).* cannot be encoded "
            r"by SignedFixedEncoder"
        )
        with pytest.raises(IllegalValue, match=pattern):
            encoder(value)
        return

    lower, upper = compute_signed_fixed_bounds(value_bit_size, frac_places)
    if value < lower or value > upper:
        pattern = (
            r"Value .* cannot be encoded by SignedFixedEncoder: "
            r"Cannot be encoded in .* bits"
        )
        with pytest.raises(ValueOutOfBounds, match=pattern):
            encoder(value)
        return

    with decimal.localcontext(abi_decimal_context):
        residue = value % (TEN**-frac_places)
    if residue > 0:
        pattern = (
            r"Value .* cannot be encoded by SignedFixedEncoder: "
            r"residue .* outside allowed"
        )
        with pytest.raises(IllegalValue, match=pattern):
            encoder(value)
        return

    # Ensure no exception
    encoder(value)


# TODO: make this generic
def test_tuple_encoder():
    encoder = TupleEncoder(
        encoders=(
            UnsignedIntegerEncoder(value_bit_size=256),
            ByteStringEncoder(),
        )
    )
    expected = decode_hex(
        "0000000000000000000000000000000000000000000000000000000000000000"
        "0000000000000000000000000000000000000000000000000000000000000040"
        "0000000000000000000000000000000000000000000000000000000000000000"
    )
    actual = encoder((0, b""))
    assert actual == expected



from hypothesis import (
    given,
)

from faster_eth_abi import (
    decode,
    encode,
)
from tests.core.common.strategies import (
    uint_strs,
    uint_values,
)


@given(uint_strs, uint_values)
def test_always_positive_multi(_type, value):
    """
    Tests round trip encoding and decoding for basic types and lists of basic
    types.
    """
    assert value >= 0

    encoded_value = encode([_type], [value])
    (decoded_value,) = decode([_type], encoded_value)

    assert decoded_value >= 0



def zpad(value: bytes, length: int) -> bytes:
    return value.rjust(length, b"\x00")


def zpad32(value: bytes) -> bytes:
    return zpad(value, length=32)


def zpad_right(value: bytes, length: int) -> bytes:
    return value.ljust(length, b"\x00")


def zpad32_right(value: bytes) -> bytes:
    return zpad_right(value, length=32)


def fpad(value: bytes, length: int) -> bytes:
    return value.rjust(length, b"\xff")


def fpad32(value: bytes) -> bytes:
    return fpad(value, length=32)



from _typeshed import Incomplete
from eth_typing.abi import TypeStr as TypeStr
from faster_eth_abi.exceptions import ABITypeError as ABITypeError
from parsimonious.nodes import Node as Node
from typing import Any, Final, Generic, Literal, NoReturn, Sequence, TypeVar, final
from typing_extensions import Self

TYPE_ALIASES: Final[Incomplete]
TYPE_ALIAS_RE: Final[Incomplete]
IntSubtype: Incomplete
FixedSubtype: Incomplete
Subtype = IntSubtype | FixedSubtype
TSub = TypeVar("TSub", IntSubtype, FixedSubtype, Literal[None])

class ABIType:
    """
    Base class for results of type string parsing operations.
    """

    arrlist: Final[Incomplete]
    node: Final[Incomplete]
    def __init__(
        self, arrlist: Sequence[str] | None = None, node: Node | None = None
    ) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def to_type_str(self) -> TypeStr:
        """
        Returns the string representation of an ABI type.  This will be equal to
        the type string from which it was created.
        """

    @property
    def item_type(self) -> Self:
        """
        If this type is an array type, equal to an appropriate
        :class:`~faster_eth_abi.grammar.ABIType` instance for the array's items.
        """

    def validate(self) -> None:
        """
        Validates the properties of an ABI type against the solidity ABI spec:

        https://solidity.readthedocs.io/en/develop/abi-spec.html

        Raises :class:`~faster_eth_abi.exceptions.ABITypeError` if validation fails.
        """

    @final
    def invalidate(self, error_msg: str) -> NoReturn: ...
    @final
    @property
    def is_array(self) -> bool:
        """
        Equal to ``True`` if a type is an array type (i.e. if it has an array
        dimension list).  Otherwise, equal to ``False``.
        """

    @property
    def is_dynamic(self) -> bool:
        """
        Equal to ``True`` if a type has a dynamically sized encoding.
        Otherwise, equal to ``False``.
        """

TComp = TypeVar("TComp", bound=ABIType)

class TupleType(ABIType):
    """
    Represents the result of parsing a tuple type string e.g. "(int,bool)".
    """

    components: Final[Incomplete]
    def __init__(
        self,
        components: tuple[TComp, ...],
        arrlist: Sequence[str] | None = None,
        *,
        node: Node | None = None
    ) -> None: ...
    def to_type_str(self) -> TypeStr: ...
    @property
    def item_type(self) -> Self: ...
    def validate(self) -> None: ...
    @property
    def is_dynamic(self) -> bool: ...

class BasicType(ABIType, Generic[TSub]):
    """
    Represents the result of parsing a basic type string e.g. "uint", "address",
    "ufixed128x19[][2]".
    """

    base: Final[Incomplete]
    sub: Final[Incomplete]
    def __init__(
        self,
        base: str,
        sub: TSub | None = None,
        arrlist: Sequence | None = None,
        *,
        node: Node | None = None
    ) -> None: ...
    def to_type_str(self) -> TypeStr: ...
    @property
    def item_type(self) -> Self: ...
    @property
    def is_dynamic(self) -> bool: ...
    def validate(self) -> None: ...

BytesType = BasicType[IntSubtype]
FixedType = BasicType[FixedSubtype]

def normalize(type_str: TypeStr) -> TypeStr:
    """
    Normalizes a type string into its canonical version e.g. the type string
    'int' becomes 'int256', etc.

    :param type_str: The type string to be normalized.
    :returns: The canonical version of the input type string.
    """



"""Classes for ABI decoding logic.

Implements classes and functions for deserializing binary data into Python values
according to ABI type specifications.
"""
import abc
import decimal
from functools import (
    cached_property,
    lru_cache,
)
from types import (
    MethodType,
)
from typing import (
    Any,
    Callable,
    Final,
    Generic,
    Optional,
    Tuple,
    TypeVar,
    Union,
    final,
)

from eth_typing import (
    HexAddress,
)
from faster_eth_utils import (
    big_endian_to_int,
    to_normalized_address,
)

from faster_eth_abi._decoding import (
    decode_dynamic_array,
    decode_head_tail,
    decode_sized_array,
    decode_tuple,
    decoder_fn_boolean,
    get_value_byte_size,
    read_fixed_byte_size_data_from_stream,
    split_data_and_padding_fixed_byte_size,
    validate_padding_bytes_fixed_byte_size,
    validate_padding_bytes_signed_integer,
    validate_pointers_array,
)
from faster_eth_abi.base import (
    BaseCoder,
)
from faster_eth_abi.exceptions import (
    InsufficientDataBytes,
    NonEmptyPaddingBytes,
)
from faster_eth_abi.from_type_str import (
    parse_tuple_type_str,
    parse_type_str,
)
from faster_eth_abi.io import (
    ContextFramesBytesIO,
)
from faster_eth_abi.typing import (
    T,
)
from faster_eth_abi.utils.numeric import (
    TEN,
    abi_decimal_context,
    ceil32,
)

TByteStr = TypeVar("TByteStr", bytes, str)


class BaseDecoder(BaseCoder, Generic[T], metaclass=abc.ABCMeta):
    """
    Base class for all decoder classes.  Subclass this if you want to define a
    custom decoder class.  Subclasses must also implement
    :any:`BaseCoder.from_type_str`.
    """

    strict = True

    @abc.abstractmethod
    def decode(self, stream: ContextFramesBytesIO) -> T:  # pragma: no cover
        """
        Decodes the given stream of bytes into a python value.  Should raise
        :any:`exceptions.DecodingError` if a python value cannot be decoded
        from the given byte stream.
        """

    def __call__(self, stream: ContextFramesBytesIO) -> T:
        return self.decode(stream)


class HeadTailDecoder(BaseDecoder[T]):
    """
    Decoder for a dynamic element of a dynamic container (a dynamic array, or a sized
    array or tuple that contains dynamic elements). A dynamic element consists of a
    pointer, aka offset, which is located in the head section of the encoded container,
    and the actual value, which is located in the tail section of the encoding.
    """

    is_dynamic = True

    def __init__(
        self,
        tail_decoder: Union[  # type: ignore [type-var]
            "HeadTailDecoder[T]",
            "SizedArrayDecoder[T]",
            "DynamicArrayDecoder[T]",
            "ByteStringDecoder[T]",
        ],
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        if tail_decoder is None:
            raise ValueError("No `tail_decoder` set")

        self.tail_decoder: Final = tail_decoder

    def decode(self, stream: ContextFramesBytesIO) -> T:
        return decode_head_tail(self, stream)

    __call__ = decode


class TupleDecoder(BaseDecoder[Tuple[T, ...]]):
    decoders: Tuple[BaseDecoder[T], ...] = ()

    def __init__(self, decoders: Tuple[BaseDecoder[T], ...], **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self.decoders = decoders = tuple(
            HeadTailDecoder(tail_decoder=d) if getattr(d, "is_dynamic", False) else d
            for d in decoders
        )

        self.is_dynamic = any(getattr(d, "is_dynamic", False) for d in decoders)
        self.len_of_head = sum(
            getattr(decoder, "array_size", 1) for decoder in decoders
        )
        self._is_head_tail = tuple(
            isinstance(decoder, HeadTailDecoder) for decoder in decoders
        )
        self._no_head_tail = not any(self._is_head_tail)

    def validate(self) -> None:
        super().validate()

        if self.decoders is None:
            raise ValueError("No `decoders` set")

    @final
    def validate_pointers(self, stream: ContextFramesBytesIO) -> None:
        raise NotImplementedError("didnt call __init__")

    def decode(self, stream: ContextFramesBytesIO) -> Tuple[T, ...]:
        return decode_tuple(self, stream)

    __call__ = decode

    @parse_tuple_type_str
    def from_type_str(cls, abi_type, registry):
        decoders = tuple(
            registry.get_decoder(c.to_type_str()) for c in abi_type.components
        )

        return cls(decoders=decoders)


class SingleDecoder(BaseDecoder[T]):
    decoder_fn = None

    def validate(self) -> None:
        super().validate()

        if self.decoder_fn is None:
            raise ValueError("No `decoder_fn` set")

    def validate_padding_bytes(self, value: Any, padding_bytes: bytes) -> None:
        raise NotImplementedError("Must be implemented by subclasses")

    def decode(self, stream: ContextFramesBytesIO) -> T:
        raw_data = self.read_data_from_stream(stream)
        data, padding_bytes = self.split_data_and_padding(raw_data)
        value = self.decoder_fn(data)  # type: ignore [misc]
        self.validate_padding_bytes(value, padding_bytes)

        return value

    __call__ = decode

    def read_data_from_stream(self, stream: ContextFramesBytesIO) -> bytes:
        raise NotImplementedError("Must be implemented by subclasses")

    def split_data_and_padding(self, raw_data: bytes) -> Tuple[bytes, bytes]:
        return raw_data, b""


class BaseArrayDecoder(BaseDecoder[Tuple[T, ...]]):
    item_decoder: BaseDecoder = None

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        # Use a head-tail decoder to decode dynamic elements
        item_decoder = self.item_decoder
        if item_decoder.is_dynamic:
            self.item_decoder = HeadTailDecoder(tail_decoder=item_decoder)
            self.validate_pointers = MethodType(validate_pointers_array, self)
        else:

            def noop(stream: ContextFramesBytesIO, array_size: int) -> None:
                ...

            self.validate_pointers = noop

    def decode(self, stream: ContextFramesBytesIO) -> Tuple[T, ...]:
        raise NotImplementedError  # this is a type stub

    def validate(self) -> None:
        super().validate()

        if self.item_decoder is None:
            raise ValueError("No `item_decoder` set")

    @parse_type_str(with_arrlist=True)
    def from_type_str(cls, abi_type, registry):
        item_decoder = registry.get_decoder(abi_type.item_type.to_type_str())

        array_spec = abi_type.arrlist[-1]
        if len(array_spec) == 1:
            # If array dimension is fixed
            return SizedArrayDecoder(
                array_size=array_spec[0],
                item_decoder=item_decoder,
            )
        else:
            # If array dimension is dynamic
            return DynamicArrayDecoder(item_decoder=item_decoder)

    def validate_pointers(self, stream: ContextFramesBytesIO, array_size: int) -> None:
        """
        Verify that all pointers point to a valid location in the stream.
        """
        validate_pointers_array(self, stream, array_size)


class SizedArrayDecoder(BaseArrayDecoder[T]):
    array_size: int = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.is_dynamic = self.item_decoder.is_dynamic

    def decode(self, stream: ContextFramesBytesIO) -> Tuple[T, ...]:
        return decode_sized_array(self, stream)

    __call__ = decode


class DynamicArrayDecoder(BaseArrayDecoder[T]):
    # Dynamic arrays are always dynamic, regardless of their elements
    is_dynamic = True

    def decode(self, stream: ContextFramesBytesIO) -> Tuple[T, ...]:
        return decode_dynamic_array(self, stream)

    __call__ = decode


class FixedByteSizeDecoder(SingleDecoder[T]):
    decoder_fn: Callable[[bytes], T] = None
    value_bit_size: int = None
    data_byte_size: int = None
    is_big_endian: bool = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.read_data_from_stream = MethodType(
            read_fixed_byte_size_data_from_stream, self
        )
        self.split_data_and_padding = MethodType(
            split_data_and_padding_fixed_byte_size, self
        )
        self._get_value_byte_size = MethodType(get_value_byte_size, self)

        # Only assign validate_padding_bytes if not overridden in subclass
        if type(self).validate_padding_bytes is SingleDecoder.validate_padding_bytes:
            self.validate_padding_bytes = MethodType(
                validate_padding_bytes_fixed_byte_size, self
            )

    def validate(self) -> None:
        super().validate()

        value_bit_size = self.value_bit_size
        if value_bit_size is None:
            raise ValueError("`value_bit_size` may not be None")
        data_byte_size = self.data_byte_size
        if data_byte_size is None:
            raise ValueError("`data_byte_size` may not be None")
        if self.decoder_fn is None:
            raise ValueError("`decoder_fn` may not be None")
        if self.is_big_endian is None:
            raise ValueError("`is_big_endian` may not be None")

        if value_bit_size % 8 != 0:
            raise ValueError(
                f"Invalid value bit size: {value_bit_size}. Must be a multiple of 8"
            )

        if value_bit_size > data_byte_size * 8:
            raise ValueError("Value byte size exceeds data size")

    def read_data_from_stream(self, stream: ContextFramesBytesIO) -> bytes:
        raise NotImplementedError("didnt call __init__")

    def split_data_and_padding(self, raw_data: bytes) -> Tuple[bytes, bytes]:
        raise NotImplementedError("didnt call __init__")

    # This is unused, but it is kept in to preserve the eth-abi api
    def _get_value_byte_size(self) -> int:
        raise NotImplementedError("didnt call __init__")


class Fixed32ByteSizeDecoder(FixedByteSizeDecoder[T]):
    data_byte_size = 32


class BooleanDecoder(Fixed32ByteSizeDecoder[bool]):
    value_bit_size = 8
    is_big_endian = True

    decoder_fn = staticmethod(decoder_fn_boolean)

    @parse_type_str("bool")
    def from_type_str(cls, abi_type, registry):
        return cls()


class AddressDecoder(Fixed32ByteSizeDecoder[HexAddress]):
    value_bit_size = 20 * 8
    is_big_endian = True
    decoder_fn = staticmethod(to_normalized_address)

    @parse_type_str("address")
    def from_type_str(cls, abi_type, registry):
        return cls()


#
# Unsigned Integer Decoders
#
class UnsignedIntegerDecoder(Fixed32ByteSizeDecoder[int]):
    decoder_fn: "staticmethod[[bytes], int]" = staticmethod(big_endian_to_int)
    is_big_endian = True

    @parse_type_str("uint")
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub)


decode_uint_256 = UnsignedIntegerDecoder(value_bit_size=256)


class UnsignedIntegerDecoderCached(UnsignedIntegerDecoder):
    decoder_fn: Callable[[bytes], int]
    maxsize: Final[Optional[int]]

    def __init__(self, maxsize: Optional[int] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.maxsize = maxsize
        self.decoder_fn = lru_cache(maxsize=maxsize)(self.decoder_fn)


#
# Signed Integer Decoders
#
class SignedIntegerDecoder(Fixed32ByteSizeDecoder[int]):
    is_big_endian = True

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        # Only assign validate_padding_bytes if not overridden in subclass
        if (
            type(self).validate_padding_bytes
            is SignedIntegerDecoder.validate_padding_bytes
        ):
            self.validate_padding_bytes = MethodType(
                validate_padding_bytes_signed_integer, self
            )

    @cached_property
    def neg_threshold(self) -> int:
        return int(2 ** (self.value_bit_size - 1))

    @cached_property
    def neg_offset(self) -> int:
        return int(2**self.value_bit_size)

    def decoder_fn(self, data: bytes) -> int:
        value = big_endian_to_int(data)
        if value >= self.neg_threshold:
            value -= self.neg_offset
        return value

    def validate_padding_bytes(self, value: Any, padding_bytes: bytes) -> None:
        return validate_padding_bytes_signed_integer(self, value, padding_bytes)

    @parse_type_str("int")
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub)


class SignedIntegerDecoderCached(SignedIntegerDecoder):
    decoder_fn: Callable[[bytes], int]
    maxsize: Final[Optional[int]]

    def __init__(self, maxsize: Optional[int] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.maxsize = maxsize
        self.decoder_fn = lru_cache(maxsize=maxsize)(self.decoder_fn)


#
# Bytes1..32
#
class BytesDecoder(Fixed32ByteSizeDecoder[bytes]):
    is_big_endian = False

    @staticmethod
    def decoder_fn(data: bytes) -> bytes:
        return data

    @parse_type_str("bytes")
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub * 8)


class BaseFixedDecoder(Fixed32ByteSizeDecoder[decimal.Decimal]):
    frac_places: int = None
    is_big_endian = True

    @cached_property
    def denominator(self) -> decimal.Decimal:
        return TEN**self.frac_places

    def validate(self) -> None:
        super().validate()

        frac_places = self.frac_places
        if frac_places is None:
            raise ValueError("must specify `frac_places`")

        if frac_places <= 0 or frac_places > 80:
            raise ValueError("`frac_places` must be in range (0, 80)")


class UnsignedFixedDecoder(BaseFixedDecoder):
    def decoder_fn(self, data: bytes) -> decimal.Decimal:
        value = big_endian_to_int(data)

        with decimal.localcontext(abi_decimal_context):
            decimal_value = decimal.Decimal(value) / self.denominator

        return decimal_value

    @parse_type_str("ufixed")
    def from_type_str(cls, abi_type, registry):
        value_bit_size, frac_places = abi_type.sub

        return cls(value_bit_size=value_bit_size, frac_places=frac_places)


class SignedFixedDecoder(BaseFixedDecoder):
    @cached_property
    def neg_threshold(self) -> int:
        return int(2 ** (self.value_bit_size - 1))

    @cached_property
    def neg_offset(self) -> int:
        return int(2**self.value_bit_size)

    @cached_property
    def expected_padding_pos(self) -> bytes:
        value_byte_size = get_value_byte_size(self)
        padding_size = self.data_byte_size - value_byte_size
        return b"\x00" * padding_size

    @cached_property
    def expected_padding_neg(self) -> bytes:
        value_byte_size = get_value_byte_size(self)
        padding_size = self.data_byte_size - value_byte_size
        return b"\xff" * padding_size

    def decoder_fn(self, data: bytes) -> decimal.Decimal:
        value = big_endian_to_int(data)
        if value >= self.neg_threshold:
            value -= self.neg_offset

        with decimal.localcontext(abi_decimal_context):
            decimal_value = decimal.Decimal(value) / self.denominator

        return decimal_value

    def validate_padding_bytes(self, value: Any, padding_bytes: bytes) -> None:
        if value >= 0:
            expected_padding_bytes = self.expected_padding_pos
        else:
            expected_padding_bytes = self.expected_padding_neg

        if padding_bytes != expected_padding_bytes:
            raise NonEmptyPaddingBytes(
                f"Padding bytes were not empty: {padding_bytes!r}"
            )

    @parse_type_str("fixed")
    def from_type_str(cls, abi_type, registry):
        value_bit_size, frac_places = abi_type.sub

        return cls(value_bit_size=value_bit_size, frac_places=frac_places)


#
# String and Bytes
#
class ByteStringDecoder(SingleDecoder[TByteStr]):
    is_dynamic = True

    @staticmethod
    def decoder_fn(data: bytes) -> bytes:
        return data

    def read_data_from_stream(self, stream: ContextFramesBytesIO) -> bytes:
        data_length = decode_uint_256(stream)
        padded_length = ceil32(data_length)

        data = stream.read(padded_length)

        if self.strict:
            if len(data) < padded_length:
                raise InsufficientDataBytes(
                    f"Tried to read {padded_length} bytes, only got {len(data)} bytes"
                )

            padding_bytes = data[data_length:]
            if padding_bytes != b"\x00" * (padded_length - data_length):
                raise NonEmptyPaddingBytes(
                    f"Padding bytes were not empty: {padding_bytes!r}"
                )

        return data[:data_length]

    def validate_padding_bytes(self, value: Any, padding_bytes: bytes) -> None:
        pass

    @parse_type_str("bytes")
    def from_type_str(cls, abi_type, registry):
        return cls()


class StringDecoder(ByteStringDecoder[str]):
    def __init__(self, handle_string_errors: str = "strict") -> None:
        self.bytes_errors: Final = handle_string_errors
        super().__init__()

    @parse_type_str("string")
    def from_type_str(cls, abi_type, registry):
        return cls()

    def decode(self, stream: ContextFramesBytesIO) -> str:
        raw_data = self.read_data_from_stream(stream)
        data, padding_bytes = self.split_data_and_padding(raw_data)
        return self.decoder_fn(data, self.bytes_errors)

    __call__ = decode

    @staticmethod
    def decoder_fn(data: bytes, handle_string_errors: str = "strict") -> str:
        return data.decode("utf-8", errors=handle_string_errors)



import random

from faster_eth_utils import (
    encode_hex,
)
import hypothesis.strategies as st

from faster_eth_abi.utils.numeric import (
    scale_places,
)

total_bits = st.integers(min_value=1, max_value=32).map(lambda n: n * 8)
frac_places = st.integers(min_value=1, max_value=80)
bytes_sizes = st.integers(min_value=1, max_value=32)

fixed_sizes = st.tuples(total_bits, frac_places)


##########################
# Type string strategies #
##########################


def join(xs):
    return "".join(map(str, xs))


def join_with_x(xs):
    return "x".join(map(str, xs))


bare_type_strs = st.sampled_from(
    [
        "uint",
        "int",
        "ufixed",
        "fixed",
        "address",
        "bool",
        "bytes",
        "function",
        "string",
    ]
)

fixed_bytes_type_strs = bytes_sizes.map(lambda x: f"bytes{x}")
uint_type_strs = total_bits.map(lambda x: f"uint{x}")
int_type_strs = total_bits.map(lambda x: f"int{x}")

fixed_size_strs = fixed_sizes.map(join_with_x)
ufixed_type_strs = fixed_size_strs.map(lambda x: f"ufixed{x}")
fixed_type_strs = fixed_size_strs.map(lambda x: f"fixed{x}")

non_array_type_strs = st.one_of(
    bare_type_strs,
    fixed_bytes_type_strs,
    uint_type_strs,
    int_type_strs,
    ufixed_type_strs,
    fixed_type_strs,
)

dynam_array_components = st.just(tuple())
fixed_array_components = st.integers(min_value=1).map(lambda x: (x,))
array_components = st.one_of(dynam_array_components, fixed_array_components)

array_lists = st.lists(array_components, min_size=1, max_size=6)
array_list_strs = array_lists.map(lambda x: "".join(repr(list(i)) for i in x))

array_type_strs = st.tuples(non_array_type_strs, array_list_strs).map(join)

non_tuple_type_strs = st.one_of(non_array_type_strs, array_type_strs)


def join_tuple(xs):
    if not isinstance(xs, list):
        return xs

    return f"({','.join(join_tuple(x) for x in xs)})"


tuple_type_strs = st.recursive(
    st.lists(non_tuple_type_strs, min_size=1, max_size=10),
    lambda this_strategy: st.lists(
        st.one_of(non_tuple_type_strs, this_strategy),
        min_size=1,
        max_size=10,
    ),
).map(join_tuple)

type_strs = st.one_of(non_tuple_type_strs, tuple_type_strs)


def guaranteed_permute(xs):
    len_xs = len(xs)
    indices = tuple(range(len_xs))

    shuffled_indices = indices
    while indices == shuffled_indices:
        shuffled_indices = tuple(random.sample(indices, k=len_xs))

    return tuple(xs[i] for i in shuffled_indices)


malformed_non_tuple_type_strs = (
    st.tuples(
        st.one_of(bare_type_strs, st.text()),
        st.one_of(total_bits, fixed_size_strs),
        array_list_strs,
    )
    .map(guaranteed_permute)
    .map(join)
)

malformed_tuple_type_strs = st.recursive(
    st.lists(malformed_non_tuple_type_strs, min_size=1, max_size=10),
    lambda this_strategy: st.lists(
        st.one_of(malformed_non_tuple_type_strs, this_strategy),
        min_size=1,
        max_size=10,
    ),
    max_leaves=5,
).map(join_tuple)

malformed_type_strs = st.one_of(
    malformed_non_tuple_type_strs,
    malformed_tuple_type_strs,
)


#################################
# Type string w/data strategies #
#################################

MIN_LIST_SIZE = 1
MAX_LIST_SIZE = 8

uint_total_bits = st.shared(total_bits, key="uint_total_bits")
uint_strs = uint_total_bits.map(lambda x: f"uint{x}")
uint_values = uint_total_bits.flatmap(
    lambda n: st.integers(
        min_value=0,
        max_value=2**n - 1,
    )
)

int_total_bits = st.shared(total_bits, key="int_total_bits")
int_strs = int_total_bits.map(lambda x: f"int{x}")
int_values = int_total_bits.flatmap(
    lambda n: st.integers(
        min_value=-(2 ** (n - 1)),
        max_value=2 ** (n - 1) - 1,
    )
)

ufixed_size_tuples = st.shared(fixed_sizes, key="ufixed_size_tuples")
ufixed_strs = ufixed_size_tuples.map(join_with_x).map(lambda x: f"ufixed{x}")
ufixed_values = ufixed_size_tuples.flatmap(
    lambda sz: st.decimals(
        min_value=0,
        max_value=2 ** sz[0] - 1,
        places=0,
    ).map(scale_places(sz[1]))
)

fixed_size_tuples = st.shared(fixed_sizes, key="fixed_size_tuples")
fixed_strs = fixed_size_tuples.map(join_with_x).map(lambda x: f"fixed{x}")
fixed_values = fixed_size_tuples.flatmap(
    lambda sz: st.decimals(
        min_value=-(2 ** (sz[0] - 1)),
        max_value=2 ** (sz[0] - 1) - 1,
        places=0,
    ).map(scale_places(sz[1]))
)

fixed_bytes_sizes = st.shared(bytes_sizes, key="fixed_bytes_sizes")
fixed_bytes_strs = fixed_bytes_sizes.map(lambda x: f"bytes{x}")
fixed_bytes_values = fixed_bytes_sizes.flatmap(
    lambda n: st.binary(
        min_size=n,
        max_size=n,
    )
)

address_strs = st.just("address")
address_values = st.binary(min_size=20, max_size=20).map(encode_hex)

bytes_strs_and_values = st.tuples(
    st.just("bytes"),
    st.binary(min_size=0, max_size=4096),
)

non_array = (
    (uint_strs, uint_values),
    (int_strs, int_values),
    (ufixed_strs, ufixed_values),
    (fixed_strs, fixed_values),
    (fixed_bytes_strs, fixed_bytes_values),
    (address_strs, address_values),
)

non_array_strs_values = st.one_of(
    *[st.tuples(type_strs, type_values) for type_strs, type_values in non_array]
)

num_unsized_elements = st.integers(min_value=0, max_value=MAX_LIST_SIZE)
unsized_array_strs_values = num_unsized_elements.flatmap(
    lambda n: st.one_of(
        [
            st.tuples(
                type_strs.map(lambda x: f"{x}[]"),
                st.lists(type_values, min_size=n, max_size=n).map(tuple),
            )
            for type_strs, type_values in non_array
        ]
    )
)

num_sized_elements = st.integers(min_value=MIN_LIST_SIZE, max_value=MAX_LIST_SIZE)
sized_array_strs_values = num_sized_elements.flatmap(
    lambda n: st.one_of(
        [
            st.tuples(
                type_strs.map(lambda ts: f"{ts}[{n}]"),
                st.lists(type_values, min_size=n, max_size=n).map(tuple),
            )
            for type_strs, type_values in non_array
        ]
    )
)

single_strs_values = st.one_of(
    unsized_array_strs_values,
    sized_array_strs_values,
    non_array_strs_values,
    bytes_strs_and_values,
)


def to_tuple(xs):
    if not isinstance(xs, list):
        return xs

    return tuple(to_tuple(x) for x in xs)


def destructure_tuple_example(xs):
    def _recur(xs, type_strs, values):
        if not isinstance(xs, list):
            type_strs.append(xs[0])
            values.append(xs[1])
            return

        _type_strs = []
        _values = []
        for ys in xs:
            _recur(ys, _type_strs, _values)

        type_strs.append(_type_strs)
        values.append(_values)

    type_strs = []
    values = []
    _recur(xs, type_strs, values)

    return join_tuple(type_strs[0]), to_tuple(values[0])


tuple_strs_values = st.recursive(
    st.lists(single_strs_values, min_size=1, max_size=10),
    lambda this_strategy: st.lists(
        st.one_of(single_strs_values, this_strategy),
        min_size=1,
        max_size=10,
    ),
).map(destructure_tuple_example)


def unzip_strs_values(strs_values):
    type_strs, type_values = zip(*strs_values)

    return tuple(type_strs), tuple(type_values)


multi_strs_values = st.lists(
    single_strs_values,
    min_size=1,
    max_size=10,
).map(unzip_strs_values)



"""Generates Markdown summaries from pytest-benchmark diff JSON results.

Reads a diff JSON file and writes Markdown tables summarizing benchmark results for each submodule and group.
"""
import json
import os


def main():
    diff_path = "pytest_benchmark_diff.json"
    results_dir = os.path.join("benchmarks", "results")
    os.makedirs(results_dir, exist_ok=True)

    # Get repo and branch info from environment variables (for links)
    repo = os.environ.get("GITHUB_REPOSITORY", "unknown/unknown")
    branch = os.environ.get("GITHUB_HEAD_REF") or os.environ.get(
        "GITHUB_REF", "main"
    ).replace("refs/heads/", "")

    with open(diff_path, encoding="utf-8") as f:
        diff = json.load(f)

    for submodule, groupDiffs in diff.items():
        # Convert submodule name to submodule file path (e.g., faster_eth_abi.encoding -> faster_eth_abi/encoding.py)
        submoduleFile = "unknown"
        benchmarkFile = "unknown"
        m = None
        if submodule.startswith("faster_eth_abi."):
            m = submodule[len("faster_eth_abi.") :]

        if m:
            submoduleFile = f"faster_eth_abi/{m}.py"
            benchmarkFile = f"benchmarks/test_{m}_benchmarks.py"

        submoduleUrl = f"https://github.com/{repo}/blob/{branch}/{submoduleFile}"
        benchmarkUrl = f"https://github.com/{repo}/blob/{branch}/{benchmarkFile}"

        md_lines = []
        md_lines.append(
            f"#### [{submodule}]({submoduleUrl}) - [view benchmarks]({benchmarkUrl})\n"
        )
        md_lines.append(
            "| Function | Reference Mean | Faster Mean | % Change | Speedup (%) | x Faster | Faster |"
        )
        md_lines.append(
            "|----------|---------------|-------------|----------|-------------|----------|--------|"
        )

        for group, data in sorted(groupDiffs.items()):
            if data.get("percent_change") is not None:
                emoji = "➖"
                if data["percent_change"] > 0:
                    emoji = "✅"
                elif data["percent_change"] < 0:
                    emoji = "❌"
                percentChange = (
                    f"{data['percent_change']:.2f}%"
                    if data.get("percent_change") is not None
                    else ""
                )
                speedupPercent = (
                    f"{data['speedup_percent']:.2f}%"
                    if data.get("speedup_percent") is not None
                    else ""
                )
                speedupX = (
                    f"{data['speedup_x']:.2f}x"
                    if data.get("speedup_x") is not None
                    and isinstance(data["speedup_x"], (int, float))
                    else ""
                )
                md_lines.append(
                    f"| `{group}` | {data.get('reference_mean', '')} | {data.get('faster_mean', '')} | {percentChange} | {speedupPercent} | {speedupX} | {emoji} |"
                )
            elif data.get("note"):
                md_lines.append(f"| `{group}` |  |  |  |  |  | ➖ |")

        md_lines.append("")  # Blank line at end
        md_content = "\n".join(md_lines)

        # Write to file
        module_name = submodule.split(".")[-1]
        out_path = os.path.join(results_dir, f"{module_name}.md")
        with open(out_path, "w", encoding="utf-8") as outf:
            outf.write(md_content)


if __name__ == "__main__":
    main()



"""Private helpers for encoding logic, intended for C compilation.

This file exists because the original encoding.py is not ready to be fully compiled to C.
This module contains functions and logic that we do wish to compile.
"""
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from faster_eth_utils import (
    is_list_like,
)

from faster_eth_abi.exceptions import (
    ValueOutOfBounds,
)

if TYPE_CHECKING:
    from faster_eth_abi.encoding import (
        BaseEncoder,
        TupleEncoder,
    )


T = TypeVar("T")


# TupleEncoder
def validate_tuple(self: "TupleEncoder", value: Sequence[Any]) -> None:
    # if we check list and tuple first it compiles to much quicker C code
    if not isinstance(value, (list, tuple)) and not is_list_like(value):
        self.invalidate_value(
            value,
            msg="must be list-like object such as array or tuple",
        )

    validators = self.validators
    if len(value) != len(validators):
        self.invalidate_value(
            value,
            exc=ValueOutOfBounds,
            msg=f"value has {len(value)} items when {len(validators)} " "were expected",
        )

    for item, validator in zip(value, validators):
        validator(item)


def encode_tuple(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    raw_head_chunks: List[Optional[bytes]] = []
    tail_chunks: List[bytes] = []
    for value, encoder, is_dynamic in zip(values, self.encoders, self._is_dynamic):
        if is_dynamic:
            raw_head_chunks.append(None)
            tail_chunks.append(encoder(value))
        else:
            raw_head_chunks.append(encoder(value))
            tail_chunks.append(b"")

    head_length = sum(32 if item is None else len(item) for item in raw_head_chunks)
    tail_offsets = [0]
    total_offset = 0
    for item in tail_chunks[:-1]:
        total_offset += len(item)
        tail_offsets.append(total_offset)

    head_chunks = tuple(
        encode_uint_256(head_length + offset) if chunk is None else chunk
        for chunk, offset in zip(raw_head_chunks, tail_offsets)
    )

    return b"".join(head_chunks) + b"".join(tail_chunks)


def encode_tuple_all_dynamic(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    tail_chunks = [encoder(value) for encoder, value in zip(encoders, values)]

    total_offset = 0
    head_length = 32 * len(encoders)
    head_chunks = [encode_uint_256(head_length)]
    for item in tail_chunks[:-1]:
        total_offset += len(item)
        head_chunks.append(encode_uint_256(head_length + total_offset))

    return b"".join(head_chunks) + b"".join(tail_chunks)


def encode_tuple_no_dynamic(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(len(encoders)))


def encode_tuple_no_dynamic1(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders: Tuple["BaseEncoder"] = self.encoders
    return encoders[0](values[0])


def encode_tuple_no_dynamic2(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder"] = self.encoders
    return encoders[0](values[0]) + encoders[1](values[1])


def encode_tuple_no_dynamic3(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder", "BaseEncoder"] = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(3))


def encode_tuple_no_dynamic4(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder"] = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(4))


def encode_tuple_no_dynamic5(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder"] = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(5))


def encode_tuple_no_dynamic6(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder"] = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(6))


def encode_tuple_no_dynamic7(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder"] = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(7))


def encode_tuple_no_dynamic8(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder"] = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(8))


def encode_tuple_no_dynamic9(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder"] = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(9))


def encode_tuple_no_dynamic10(self: "TupleEncoder", values: Sequence[Any]) -> bytes:
    validate_tuple(self, values)
    encoders = self.encoders
    # encoders: Tuple["BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder", "BaseEncoder"] = self.encoders
    return b"".join(encoders[i](values[i]) for i in range(10))


encode_tuple_no_dynamic_funcs: Dict[
    int, Callable[["TupleEncoder", Sequence[Any]], bytes]
] = {
    1: encode_tuple_no_dynamic1,
    2: encode_tuple_no_dynamic2,
    3: encode_tuple_no_dynamic3,
    4: encode_tuple_no_dynamic4,
    5: encode_tuple_no_dynamic5,
    6: encode_tuple_no_dynamic6,
    7: encode_tuple_no_dynamic7,
    8: encode_tuple_no_dynamic8,
    9: encode_tuple_no_dynamic9,
    10: encode_tuple_no_dynamic10,
}


def encode_fixed(
    value: Any,
    encode_fn: Callable[[Any], bytes],
    is_big_endian: bool,
    data_byte_size: int,
) -> bytes:
    base_encoded_value = encode_fn(value)
    if is_big_endian:
        return base_encoded_value.rjust(data_byte_size, b"\x00")
    else:
        return base_encoded_value.ljust(data_byte_size, b"\x00")


def encode_signed(
    value: T,
    encode_fn: Callable[[T], bytes],
    data_byte_size: int,
) -> bytes:
    base_encoded_value = encode_fn(value)
    if value >= 0:
        return base_encoded_value.rjust(data_byte_size, b"\x00")
    else:
        return base_encoded_value.rjust(data_byte_size, b"\xff")


def encode_elements(item_encoder: "BaseEncoder", value: Sequence[Any]) -> bytes:
    tail_chunks = tuple(item_encoder(i) for i in value)

    items_are_dynamic: bool = getattr(item_encoder, "is_dynamic", False)
    if not items_are_dynamic or len(value) == 0:
        return b"".join(tail_chunks)

    head_length = 32 * len(value)
    tail_offsets = [0]
    total_offset = 0
    for item in tail_chunks[:-1]:
        total_offset += len(item)
        tail_offsets.append(total_offset)

    head_chunks = tuple(
        encode_uint_256(head_length + offset) for offset in tail_offsets
    )
    return b"".join(head_chunks) + b"".join(tail_chunks)


def encode_elements_dynamic(item_encoder: "BaseEncoder", value: Sequence[Any]) -> bytes:
    encoded_size = encode_uint_256(len(value))
    encoded_elements = encode_elements(item_encoder, value)
    return encoded_size + encoded_elements


def encode_uint_256(i: int) -> bytes:
    # An optimized version of the `encode_uint_256` in `encoding.py` which
    # does not perform any validation. We should not have any issues here
    # unless you're encoding really really massive iterables.
    big_endian = int_to_big_endian(i)
    return big_endian.rjust(32, b"\x00")


def int_to_big_endian(value: int) -> bytes:
    # vendored from eth-utils so it can compile nicely into faster-eth-abi's binary
    return value.to_bytes((value.bit_length() + 7) // 8 or 1, "big")



import pytest
import re

from faster_eth_abi import (
    encode,
)
from faster_eth_abi.exceptions import (
    NoEntriesFound,
    ValueOutOfBounds,
)
from faster_eth_abi.grammar import (
    parse,
)
from tests.core.common.unit import (
    CORRECT_DYNAMIC_ENCODINGS,
    CORRECT_STATIC_ENCODINGS,
    CORRECT_TUPLE_ENCODINGS,
    words,
)


@pytest.mark.parametrize(
    "tuple_type,python_value,solidity_abi_encoded,_",
    CORRECT_TUPLE_ENCODINGS,
)
def test_abi_encode_for_multiple_types_as_list(
    tuple_type, python_value, solidity_abi_encoded, _
):
    abi_type = parse(tuple_type)
    if abi_type.arrlist is not None:
        pytest.skip("ABI coding functions do not support array types")

    # assert different types encoded correctly as a list
    # e.g. encode(['bytes32[]', 'uint256'], ([b'a', b'b'], 22))
    #
    # compare to solidity:
    #   bytes32 a = 0x6100000000000000000000000000000000000000000000000000000000000000;
    #   bytes32 b = 0x6200000000000000000000000000000000000000000000000000000000000000;
    #   bytes32[] arr = [a,b];
    #   uint256 num = 22;
    #
    #   abi.encode(arr,num);
    separated_list_of_types = [t.to_type_str() for t in abi_type.components]
    eth_abi_encoded = encode(separated_list_of_types, python_value)

    assert eth_abi_encoded == solidity_abi_encoded


@pytest.mark.parametrize(
    "single_abi_type,python_value,solidity_abi_encoded,_",
    CORRECT_STATIC_ENCODINGS,
)
def test_abi_encode_for_single_static_types(
    single_abi_type, python_value, solidity_abi_encoded, _
):
    # If single_abi_type is a tuple, assert the tuple type is encoded correctly
    # e.g. encode(['(bytes32[],uint256)'], [([b'a', b'b'], 22)])
    #
    # compare to solidity:
    #   struct TupleExample {
    #     bytes32[] arg1;
    #     uint256 arg2;
    #   }
    #   bytes32 a = 0x6100000000000000000000000000000000000000000000000000000000000000;
    #   bytes32 b = 0x6200000000000000000000000000000000000000000000000000000000000000;
    #   bytes32[] arr = [a,b];
    #   uint256 num = 22;
    #
    #   abi.encode(TupleExample(arr,num));
    eth_abi_encoded = encode([single_abi_type], [python_value])

    assert eth_abi_encoded == solidity_abi_encoded


@pytest.mark.parametrize(
    "single_abi_type,python_value,solidity_abi_encoded,_",
    CORRECT_DYNAMIC_ENCODINGS,
)
def test_abi_encode_for_single_dynamic_types(
    single_abi_type, python_value, solidity_abi_encoded, _
):
    # Same test as the single static types test above but with dynamic types
    eth_abi_encoded = encode([single_abi_type], [python_value])

    solidity_abi_encoded = (
        # 32 bytes offset for dynamic types
        b"".join([words("20"), solidity_abi_encoded])
    )

    assert eth_abi_encoded == solidity_abi_encoded


@pytest.mark.parametrize(
    "non_list_like_value",
    (
        "",
        123,
        b"",
        b"\xff",
        b"david attenborough",
        bytearray(b"\x01\xff"),
        {"key": "val"},
        {1, 2},
    ),
)
def test_abi_encode_raises_for_non_list_like_params(non_list_like_value):
    # test raises when `types` param is not list-like
    if non_list_like_value in ("", b""):
        expected_exc = ValueOutOfBounds
    elif non_list_like_value == {"key": "val"}:
        expected_exc = NoEntriesFound
    else:
        expected_exc = TypeError
    with pytest.raises(
        expected_exc,
        # NOTE The `match` arg was commented out because faster-eth-abi might not raise
        # the same exception, but we should still test that it raises an Exception like eth-abi
        # match=f"The `types` value type must be one of list or tuple. Got {type(non_list_like_value)}",
    ):
        encode(non_list_like_value, ["bytes"])

    # test raises when `args` param is not list-like
    with pytest.raises(
        TypeError,
        match=f"The `args` value type must be one of list or tuple. "
        f"Got {type(non_list_like_value)}",
    ):
        encode(["valid_string_value"], non_list_like_value)


@pytest.mark.parametrize(
    "zero_sized_tuple_type,python_value",
    (
        ("()", ()),
        ("(int,())", (1, ())),
        ("(int,((),))", (1, ((),))),
        ("(((),),int)", (((),), 1)),
        ("(((),),int,())", (((),), 1, ())),
        ("(int,(),int)", (1, (), 1)),
    ),
)
def test_abi_encode_raises_for_zero_sized_tuple_types(
    zero_sized_tuple_type,
    python_value,
):
    with pytest.raises(
        ValueError,
        match=re.escape('Zero-sized tuple types "()" are not supported.'),
    ):
        encode([zero_sized_tuple_type], [python_value])



"""Parsing and normalization logic for ABI type strings.

Implements grammar, parsing, and type validation for ABI type strings.
"""
import functools
from typing import (
    Final,
    final,
)

import parsimonious
from parsimonious import (
    expressions,
)

from faster_eth_abi._grammar import (
    TYPE_ALIAS_RE,
    TYPE_ALIASES,
    ABIType,
    BasicType,
    TupleType,
    normalize,
)
from faster_eth_abi.exceptions import (
    ParseError,
)

grammar: Final = parsimonious.Grammar(
    r"""
    type = tuple_type / basic_type

    tuple_type = components arrlist?
    components = non_zero_tuple

    non_zero_tuple = "(" type next_type* ")"
    next_type = "," type

    basic_type = base sub? arrlist?

    base = alphas

    sub = two_size / digits
    two_size = (digits "x" digits)

    arrlist = (const_arr / dynam_arr)+
    const_arr = "[" digits "]"
    dynam_arr = "[]"

    alphas = ~"[A-Za-z]+"
    digits = ~"[1-9][0-9]*"
    """
)


@final
class NodeVisitor(parsimonious.NodeVisitor):  # type: ignore [misc]
    """
    Parsimonious node visitor which performs both parsing of type strings and
    post-processing of parse trees.  Parsing operations are cached.
    """

    def __init__(self):
        self.parse = functools.lru_cache(maxsize=None)(self._parse_uncached)

    grammar = grammar

    def visit_non_zero_tuple(self, node, visited_children):
        # Ignore left and right parens
        _, first, rest, _ = visited_children

        return (first,) + rest

    def visit_tuple_type(self, node, visited_children):
        components, arrlist = visited_children

        return TupleType(components, arrlist, node=node)

    def visit_next_type(self, node, visited_children):
        # Ignore comma
        _, abi_type = visited_children

        return abi_type

    def visit_basic_type(self, node, visited_children):
        base, sub, arrlist = visited_children

        return BasicType(base, sub, arrlist, node=node)

    def visit_two_size(self, node, visited_children):
        # Ignore "x"
        first, _, second = visited_children

        return first, second

    def visit_const_arr(self, node, visited_children):
        # Ignore left and right brackets
        _, int_value, _ = visited_children

        return (int_value,)

    def visit_dynam_arr(self, node, visited_children):
        return ()

    def visit_alphas(self, node, visited_children):
        return node.text

    def visit_digits(self, node, visited_children):
        return int(node.text)

    def generic_visit(self, node, visited_children):
        expr = node.expr
        if isinstance(expr, expressions.OneOf):
            # Unwrap value chosen from alternatives
            return visited_children[0]

        if isinstance(expr, expressions.Quantifier) and expr.min == 0 and expr.max == 1:
            # Unwrap optional value or return `None`
            if len(visited_children) != 0:
                return visited_children[0]

            return None

        return tuple(visited_children)

    def _parse_uncached(self, type_str, **kwargs):
        """
        Parses a type string into an appropriate instance of
        :class:`~faster_eth_abi.grammar.ABIType`.  If a type string cannot be parsed,
        throws :class:`~faster_eth_abi.exceptions.ParseError`.

        :param type_str: The type string to be parsed.
        :returns: An instance of :class:`~faster_eth_abi.grammar.ABIType` containing
            information about the parsed type string.
        """
        if not isinstance(type_str, str):
            raise TypeError(f"Can only parse string values: got {type(type_str)}")

        try:
            return super().parse(type_str, **kwargs)
        except parsimonious.ParseError as e:
            # This is a good place to add some better messaging around the type string.
            # If this logic grows any bigger, we should abstract it to its own function.
            if "()" in type_str:
                # validate against zero-sized tuple types
                raise ValueError(
                    'Zero-sized tuple types "()" are not supported.'
                ) from None

            raise ParseError(e.text, e.pos, e.expr) from e


visitor: Final = NodeVisitor()

parse: Final = visitor.parse


__all__ = [
    "NodeVisitor",
    "ABIType",
    "TupleType",
    "BasicType",
    "grammar",
    "parse",
    "normalize",
    "visitor",
    "TYPE_ALIASES",
    "TYPE_ALIAS_RE",
]



from typing import Any

class BaseCoder:
    """
    Base class for all encoder and decoder classes.
    """

    is_dynamic: bool
    def __init__(self, **kwargs: Any) -> None: ...
    def validate(self) -> None: ...
    @classmethod
    def from_type_str(cls, type_str, registry) -> None:
        """
        Used by :any:`ABIRegistry` to get an appropriate encoder or decoder
        instance for the given type string and type registry.
        """



from .codec import ABIEncoder as ABIEncoder
from .registry import registry_packed as registry_packed
from _typeshed import Incomplete
from typing import Final

default_encoder_packed: Final[Incomplete]
encode_packed: Final[Incomplete]
is_encodable_packed: Final[Incomplete]



from typing import (
    Any,
)


def validate_bytes_param(param: Any, param_name: str) -> None:
    if not isinstance(param, (bytes, bytearray)):
        raise TypeError(
            f"The `{param_name}` value must be of bytes type. Got {type(param)}"
        )


def validate_list_like_param(param: Any, param_name: str) -> None:
    if not isinstance(param, (list, tuple)):
        raise TypeError(
            f"The `{param_name}` value type must be one of list or tuple. "
            f"Got {type(param)}"
        )



import abc
import decimal
from _typeshed import Incomplete
from faster_eth_abi._encoding import (
    encode_elements as encode_elements,
    encode_elements_dynamic as encode_elements_dynamic,
    encode_fixed as encode_fixed,
    encode_signed as encode_signed,
    encode_tuple as encode_tuple,
    encode_tuple_all_dynamic as encode_tuple_all_dynamic,
    encode_tuple_no_dynamic as encode_tuple_no_dynamic,
    encode_tuple_no_dynamic_funcs as encode_tuple_no_dynamic_funcs,
    int_to_big_endian as int_to_big_endian,
    validate_tuple as validate_tuple,
)
from faster_eth_abi.base import BaseCoder as BaseCoder
from faster_eth_abi.exceptions import (
    EncodingTypeError as EncodingTypeError,
    IllegalValue as IllegalValue,
    ValueOutOfBounds as ValueOutOfBounds,
)
from faster_eth_abi.from_type_str import (
    parse_tuple_type_str as parse_tuple_type_str,
    parse_type_str as parse_type_str,
)
from faster_eth_abi.utils.numeric import (
    TEN as TEN,
    abi_decimal_context as abi_decimal_context,
    ceil32 as ceil32,
    compute_signed_fixed_bounds as compute_signed_fixed_bounds,
    compute_signed_integer_bounds as compute_signed_integer_bounds,
    compute_unsigned_fixed_bounds as compute_unsigned_fixed_bounds,
    compute_unsigned_integer_bounds as compute_unsigned_integer_bounds,
)
from faster_eth_abi.utils.padding import zpad_right as zpad_right
from faster_eth_abi.utils.string import abbr as abbr
from functools import cached_property as cached_property
from typing import Any, Callable, ClassVar, Final, NoReturn, Sequence, final
from typing_extensions import Self

class BaseEncoder(BaseCoder, metaclass=abc.ABCMeta):
    """
    Base class for all encoder classes.  Subclass this if you want to define a
    custom encoder class.  Subclasses must also implement
    :any:`BaseCoder.from_type_str`.
    """

    @abc.abstractmethod
    def encode(self, value: Any) -> bytes:
        """
        Encodes the given value as a sequence of bytes.  Should raise
        :any:`exceptions.EncodingError` if ``value`` cannot be encoded.
        """

    @abc.abstractmethod
    def validate_value(self, value: Any) -> None:
        """
        Checks whether or not the given value can be encoded by this encoder.
        If the given value cannot be encoded, must raise
        :any:`exceptions.EncodingError`.
        """

    @classmethod
    def invalidate_value(
        cls, value: Any, exc: type[Exception] = ..., msg: str | None = None
    ) -> NoReturn:
        """
        Throws a standard exception for when a value is not encodable by an
        encoder.
        """

    def __call__(self, value: Any) -> bytes: ...

class TupleEncoder(BaseEncoder):
    encoders: tuple[BaseEncoder, ...]
    is_dynamic: Incomplete
    validators: Final[Callable[[Any], None]]
    def __init__(self, encoders: tuple[BaseEncoder, ...], **kwargs: Any) -> None: ...
    def validate(self) -> None: ...
    @final
    def validate_value(self, value: Sequence[Any]) -> None: ...
    def encode(self, values: Sequence[Any]) -> bytes: ...
    def __call__(self, values: Sequence[Any]) -> bytes: ...
    @parse_tuple_type_str
    def from_type_str(cls, abi_type, registry): ...

class FixedSizeEncoder(BaseEncoder):
    value_bit_size: Incomplete
    data_byte_size: Incomplete
    encode_fn: Incomplete
    type_check_fn: Incomplete
    is_big_endian: Incomplete
    def validate(self) -> None: ...
    def validate_value(self, value) -> None: ...
    def encode(self, value: Any) -> bytes: ...
    __call__ = encode

class Fixed32ByteSizeEncoder(FixedSizeEncoder):
    data_byte_size: int

class BooleanEncoder(Fixed32ByteSizeEncoder):
    value_bit_size: int
    is_big_endian: bool
    @classmethod
    def validate_value(cls, value: Any) -> None: ...
    @classmethod
    def encode_fn(cls, value: bool) -> bytes: ...
    def from_type_str(cls, abi_type, registry): ...

class PackedBooleanEncoder(BooleanEncoder):
    data_byte_size: int

class NumberEncoder(Fixed32ByteSizeEncoder):
    is_big_endian: bool
    bounds_fn: Incomplete
    illegal_value_fn: Incomplete
    type_check_fn: Incomplete
    def validate(self) -> None: ...
    def validate_value(self, value) -> None: ...

class UnsignedIntegerEncoder(NumberEncoder):
    encode_fn: Incomplete
    bounds_fn: Incomplete
    type_check_fn: Incomplete
    def from_type_str(cls, abi_type, registry): ...

class UnsignedIntegerEncoderCached(UnsignedIntegerEncoder):
    encode: Final[Callable[[int], bytes]]
    maxsize: Final[int | None]
    def __init__(self, maxsize: int | None = None, **kwargs: Any) -> None: ...

encode_uint_256: Incomplete

class PackedUnsignedIntegerEncoder(UnsignedIntegerEncoder):
    def from_type_str(cls, abi_type, registry): ...

class PackedUnsignedIntegerEncoderCached(PackedUnsignedIntegerEncoder):
    encode: Final[Callable[[int], bytes]]
    maxsize: Final[int | None]
    def __init__(self, maxsize: int | None = None, **kwargs: Any) -> None: ...

class SignedIntegerEncoder(NumberEncoder):
    bounds_fn: Incomplete
    type_check_fn: Incomplete
    def encode_fn(self, value: int) -> bytes: ...
    def encode(self, value: int) -> bytes: ...
    __call__ = encode
    def from_type_str(cls, abi_type, registry): ...

class SignedIntegerEncoderCached(SignedIntegerEncoder):
    encode: Final[Callable[[int], bytes]]
    maxsize: Final[int | None]
    def __init__(self, maxsize: int | None = None, **kwargs: Any) -> None: ...

class PackedSignedIntegerEncoder(SignedIntegerEncoder):
    def from_type_str(cls, abi_type, registry): ...

class PackedSignedIntegerEncoderCached(PackedSignedIntegerEncoder):
    encode: Final[Callable[[int], bytes]]
    maxsize: Final[int | None]
    def __init__(self, maxsize: int | None = None, **kwargs: Any) -> None: ...

class BaseFixedEncoder(NumberEncoder):
    frac_places: Incomplete
    @staticmethod
    def type_check_fn(value): ...
    @staticmethod
    def illegal_value_fn(value): ...
    @cached_property
    def denominator(self) -> decimal.Decimal: ...
    @cached_property
    def precision(self) -> int: ...
    def validate_value(self, value) -> None: ...
    def validate(self) -> None: ...

class UnsignedFixedEncoder(BaseFixedEncoder):
    def bounds_fn(self, value_bit_size): ...
    def encode_fn(self, value): ...
    def from_type_str(cls, abi_type, registry): ...

class PackedUnsignedFixedEncoder(UnsignedFixedEncoder):
    def from_type_str(cls, abi_type, registry): ...

class SignedFixedEncoder(BaseFixedEncoder):
    def bounds_fn(self, value_bit_size): ...
    def encode_fn(self, value): ...
    def encode(self, value): ...
    __call__ = encode
    def from_type_str(cls, abi_type, registry): ...

class PackedSignedFixedEncoder(SignedFixedEncoder):
    def from_type_str(cls, abi_type, registry): ...

class AddressEncoder(Fixed32ByteSizeEncoder):
    value_bit_size: Incomplete
    encode_fn: Incomplete
    is_big_endian: bool
    @classmethod
    def validate_value(cls, value: Any) -> None: ...
    def validate(self) -> None: ...
    def from_type_str(cls, abi_type, registry): ...

class PackedAddressEncoder(AddressEncoder):
    data_byte_size: int

class BytesEncoder(Fixed32ByteSizeEncoder):
    is_big_endian: bool
    def validate_value(self, value: Any) -> None: ...
    @staticmethod
    def encode_fn(value): ...
    def from_type_str(cls, abi_type, registry): ...

class PackedBytesEncoder(BytesEncoder):
    def from_type_str(cls, abi_type, registry): ...

class ByteStringEncoder(BaseEncoder):
    is_dynamic: bool
    @classmethod
    def validate_value(cls, value: Any) -> None: ...
    @classmethod
    def encode(cls, value: bytes) -> bytes: ...
    __call__: ClassVar[Callable[[type[Self], bytes], bytes]]
    def from_type_str(cls, abi_type, registry): ...

class PackedByteStringEncoder(ByteStringEncoder):
    is_dynamic: bool
    @classmethod
    def encode(cls, value): ...
    __call__ = encode

class TextStringEncoder(BaseEncoder):
    is_dynamic: bool
    @classmethod
    def validate_value(cls, value: Any) -> None: ...
    @classmethod
    def encode(cls, value: str) -> bytes: ...
    __call__: ClassVar[Callable[[type[Self], str], bytes]]
    def from_type_str(cls, abi_type, registry): ...

class PackedTextStringEncoder(TextStringEncoder):
    is_dynamic: bool
    @classmethod
    def encode(cls, value: str) -> bytes: ...
    __call__ = encode

class BaseArrayEncoder(BaseEncoder, metaclass=abc.ABCMeta):
    item_encoder: BaseEncoder
    def validate(self) -> None: ...
    def validate_value(self, value: Any) -> None: ...
    def encode_elements(self, value: Sequence[Any]) -> bytes: ...
    def from_type_str(cls, abi_type, registry): ...

class PackedArrayEncoder(BaseArrayEncoder):
    array_size: Incomplete
    def validate_value(self, value: Any) -> None: ...
    def encode(self, value: Sequence[Any]) -> bytes: ...
    __call__ = encode
    def from_type_str(cls, abi_type, registry): ...

class SizedArrayEncoder(BaseArrayEncoder):
    array_size: Incomplete
    is_dynamic: Incomplete
    def __init__(self, **kwargs) -> None: ...
    def validate(self) -> None: ...
    def validate_value(self, value: Any) -> None: ...
    def encode(self, value: Sequence[Any]) -> bytes: ...
    __call__ = encode

class DynamicArrayEncoder(BaseArrayEncoder):
    is_dynamic: bool
    def encode(self, value: Sequence[Any]) -> bytes: ...



from _typeshed import Incomplete
from faster_eth_abi.codec import ABICodec as ABICodec
from faster_eth_abi.registry import registry as registry
from typing import Final

default_codec: Final[Incomplete]
encode: Final[Incomplete]
decode: Final[Incomplete]
is_encodable: Final[Incomplete]
is_encodable_type: Final[Incomplete]



"""Script to autogenerate typing.py with all int/uint Literal types and custom tuple types.

This script creates the typing definitions for ABI type strings, including all integer, bytes, and tuple variants.
"""
import itertools
from pathlib import (
    Path,
)

HEADER = """# AUTOGENERATED FILE. DO NOT EDIT DIRECTLY.
from typing import (
    Literal,
    TypeVar,
    Union,
)

T = TypeVar("T")

IntegerTypeStr = Union["IntTypeStr", "UintTypeStr"]

AddressTypeStr = Literal["address"]
BoolTypeStr = Literal["bool"]
StringTypeStr = Literal["string"]

AddressArrayTypeStr = Literal["address[]"]
StringArrayTypeStr = Literal["string[]"]
BoolArrayTypeStr = Literal["bool[]"]

ArrayTypeStr = Union[
    "AddressArrayTypeStr",
    "BoolArrayTypeStr",
    "BytesArrayTypeStr",
    "IntArrayTypeStr",
    "StringArrayTypeStr",
]

TupleTypeStr = Union[
    "TupleAddressIntTypeStr",
    "TupleBoolIntTypeStr",
    "TupleBytesIntTypeStr",
    "TupleIntIntTypeStr",
    "TupleStrIntTypeStr",
]

"""

INT_SUFFIXES = range(8, 257, 8)
INT_TYPES = [f'"int{i}"' for i in INT_SUFFIXES]
UINT_TYPES = [f'"uint{i}"' for i in INT_SUFFIXES]
ALL_INT_TYPES = INT_TYPES + UINT_TYPES
BYTES_TYPES = ["bytes"] + [f"bytes{i}" for i in range(1, 33)]


def gen_int_types(prefix: str, suffix: str, int_types: list[str]) -> list[str]:
    return [f'"{prefix}{t[1:-1]}{suffix}"' for t in int_types]


def write_typing_py():
    int_types = ",\n    ".join(INT_TYPES)
    uint_types = ",\n    ".join(UINT_TYPES)

    address_int_types = ",\n    ".join(gen_int_types("(address,", ")", ALL_INT_TYPES))
    bool_int_types = ",\n    ".join(gen_int_types("(bool,", ")", ALL_INT_TYPES))
    bytes_int_types = ",\n    ".join(gen_int_types("(bytes,", ")", ALL_INT_TYPES))
    string_int_types = ",\n    ".join(gen_int_types("(string,", ")", ALL_INT_TYPES))

    bytes_types = ",\n    ".join(f'"{t}"' for t in BYTES_TYPES)
    bytes_array_types = ",\n    ".join(f'"{t}[]"' for t in BYTES_TYPES)

    int_array_types = ",\n    ".join(f'{t[:-1]}[]"' for t in ALL_INT_TYPES)
    int_int_types = ",\n    ".join(
        f'"({a[1:-1]},{b[1:-1]})"'
        for a, b in itertools.product(ALL_INT_TYPES, repeat=2)
    )

    content = HEADER
    content += f"""
IntTypeStr = Literal[
    {int_types},
]

UintTypeStr = Literal[
    {uint_types},
]

BytesTypeStr = Literal[
    {bytes_types},
]

IntArrayTypeStr = Literal[
    {int_array_types},
]

BytesArrayTypeStr = Literal[
    {bytes_array_types},
]

# New tuple type strings
TupleAddressIntTypeStr = Literal[
    {address_int_types},
]

TupleBoolIntTypeStr = Literal[
    {bool_int_types},
]

TupleBytesIntTypeStr = Literal[
    {bytes_int_types},
]

TupleIntIntTypeStr = Literal[
    {int_int_types},
]

TupleStrIntTypeStr = Literal[
    {string_int_types},
]

AnyTypeStr = Union[
    AddressTypeStr,
    BoolTypeStr,
    StringTypeStr,
    AddressArrayTypeStr,
    StringArrayTypeStr,
    BoolArrayTypeStr,
    ArrayTypeStr,
    TupleTypeStr,
    BytesTypeStr,
    IntTypeStr,
    UintTypeStr,
    str,
]
"""

    Path("faster_eth_abi/typing.py").write_text(content.strip() + "\n")


if __name__ == "__main__":
    write_typing_py()



import pytest

from faster_eth_abi.utils.padding import (
    zpad,
)


@pytest.mark.parametrize(
    "value,length,expected",
    (
        (b"", 5, b"\x00\x00\x00\x00\x00"),
        (b"abc", 5, b"\x00\x00abc"),
        (b"abcde", 5, b"abcde"),
        (b"abcdef", 5, b"abcdef"),
    ),
)
def test_zpadding(value, length, expected):
    actual = zpad(value, length)
    assert actual == expected



def zpad(value: bytes, length: int) -> bytes: ...
def zpad32(value: bytes) -> bytes: ...
def zpad_right(value: bytes, length: int) -> bytes: ...
def zpad32_right(value: bytes) -> bytes: ...
def fpad(value: bytes, length: int) -> bytes: ...
def fpad32(value: bytes) -> bytes: ...



from typing import (
    Final,
)

from faster_eth_abi.codec import (
    ABICodec,
)
from faster_eth_abi.registry import (
    registry,
)

default_codec: Final = ABICodec(registry)

encode: Final = default_codec.encode
decode: Final = default_codec.decode
is_encodable: Final = default_codec.is_encodable
is_encodable_type: Final = default_codec.is_encodable_type



# AUTOGENERATED: Run scripts/generate_overload_tests.py to update.
# Requires mypy extras
# To run: pip install mypy && mypy tests/typecheck

# alias some things so we can minimize test data size in source control
from typing import Any as A
from typing import Tuple as T
from typing import Union as U
from typing_extensions import assert_type as ck
from eth_typing import HexAddress as HA
from faster_eth_abi.codec import ABIDecoder
from faster_eth_abi.registry import ABIRegistry

# more aliases to save a bunch more chars
bo = bool
by = bytes
i = int
s = str

TA = T[A, ...]

THA = T[HA, ...]
Tbo = T[bo, ...]
Tby = T[by, ...]
Ti = T[i, ...]
Ts = T[s, ...]
TUa_bo = T[U[HA, bo], ...]
TUa_by = T[U[HA, by], ...]
TUa_i = T[U[HA, i], ...]
TUa_s = T[U[HA, s], ...]
TUbo_by = T[U[bo, by], ...]
TUbo_i = T[U[bo, i], ...]
TUbo_s = T[U[bo, s], ...]
TUby_i = T[U[by, i], ...]
TUby_s = T[U[by, s], ...]
TUi_s = T[U[i, s], ...]
TUa_bo_by = T[U[bo, by, HA], ...]
TUbo_by_i = T[U[bo, by, i], ...]
TUbo_by_s = T[U[bo, by, s], ...]
TUa_bo_i = T[U[bo, i, HA], ...]
TUa_by_i = T[U[by, i, HA], ...]
TUby_i_s = T[U[by, i, s], ...]

decoder = ABIDecoder(ABIRegistry())

# the callable we are checking
f = decoder.decode

# input data
D = b"\x00" * 32
ck(f(('address',), D), T[HA])  # fixed case 0
ck(f(('address[]',), D), T[A])  # fixed case 1
ck(f(('bool',), D), T[bo])  # fixed case 2
ck(f(('bool[]',), D), T[A])  # fixed case 3
ck(f(('bytes',), D), T[by])  # fixed case 4
ck(f(('bytes1',), D), T[by])  # fixed case 5
ck(f(('bytes10',), D), T[by])  # fixed case 6
ck(f(('bytes11',), D), T[by])  # fixed case 7
ck(f(('bytes12',), D), T[by])  # fixed case 8
ck(f(('bytes13',), D), T[by])  # fixed case 9
ck(f(('bytes14',), D), T[by])  # fixed case 10
ck(f(('bytes15',), D), T[by])  # fixed case 11
ck(f(('bytes16',), D), T[by])  # fixed case 12
ck(f(('bytes17',), D), T[by])  # fixed case 13
ck(f(('bytes18',), D), T[by])  # fixed case 14
ck(f(('bytes19',), D), T[by])  # fixed case 15
ck(f(('bytes2',), D), T[by])  # fixed case 16
ck(f(('bytes20',), D), T[by])  # fixed case 17
ck(f(('bytes21',), D), T[by])  # fixed case 18
ck(f(('bytes22',), D), T[by])  # fixed case 19
ck(f(('bytes23',), D), T[by])  # fixed case 20
ck(f(('bytes24',), D), T[by])  # fixed case 21
ck(f(('bytes25',), D), T[by])  # fixed case 22
ck(f(('bytes26',), D), T[by])  # fixed case 23
ck(f(('bytes27',), D), T[by])  # fixed case 24
ck(f(('bytes28',), D), T[by])  # fixed case 25
ck(f(('bytes29',), D), T[by])  # fixed case 26
ck(f(('bytes3',), D), T[by])  # fixed case 27
ck(f(('bytes30',), D), T[by])  # fixed case 28
ck(f(('bytes31',), D), T[by])  # fixed case 29
ck(f(('bytes32',), D), T[by])  # fixed case 30
ck(f(('bytes4',), D), T[by])  # fixed case 31
ck(f(('bytes5',), D), T[by])  # fixed case 32
ck(f(('bytes6',), D), T[by])  # fixed case 33
ck(f(('bytes7',), D), T[by])  # fixed case 34
ck(f(('bytes8',), D), T[by])  # fixed case 35
ck(f(('bytes9',), D), T[by])  # fixed case 36
ck(f(('int104',), D), T[i])  # fixed case 37
ck(f(('int112',), D), T[i])  # fixed case 38
ck(f(('int120',), D), T[i])  # fixed case 39
ck(f(('int128',), D), T[i])  # fixed case 40
ck(f(('int136',), D), T[i])  # fixed case 41
ck(f(('int144',), D), T[i])  # fixed case 42
ck(f(('int152',), D), T[i])  # fixed case 43
ck(f(('int16',), D), T[i])  # fixed case 44
ck(f(('int160',), D), T[i])  # fixed case 45
ck(f(('int168',), D), T[i])  # fixed case 46
ck(f(('int176',), D), T[i])  # fixed case 47
ck(f(('int184',), D), T[i])  # fixed case 48
ck(f(('int192',), D), T[i])  # fixed case 49
ck(f(('int200',), D), T[i])  # fixed case 50
ck(f(('int208',), D), T[i])  # fixed case 51
ck(f(('int216',), D), T[i])  # fixed case 52
ck(f(('int224',), D), T[i])  # fixed case 53
ck(f(('int232',), D), T[i])  # fixed case 54
ck(f(('int24',), D), T[i])  # fixed case 55
ck(f(('int240',), D), T[i])  # fixed case 56
ck(f(('int248',), D), T[i])  # fixed case 57
ck(f(('int256',), D), T[i])  # fixed case 58
ck(f(('int32',), D), T[i])  # fixed case 59
ck(f(('int40',), D), T[i])  # fixed case 60
ck(f(('int48',), D), T[i])  # fixed case 61
ck(f(('int56',), D), T[i])  # fixed case 62
ck(f(('int64',), D), T[i])  # fixed case 63
ck(f(('int72',), D), T[i])  # fixed case 64
ck(f(('int8',), D), T[i])  # fixed case 65
ck(f(('int80',), D), T[i])  # fixed case 66
ck(f(('int88',), D), T[i])  # fixed case 67
ck(f(('int96',), D), T[i])  # fixed case 68
ck(f(('string',), D), T[s])  # fixed case 69
ck(f(('string[]',), D), T[A])  # fixed case 70
ck(f(('uint104',), D), T[i])  # fixed case 71
ck(f(('uint112',), D), T[i])  # fixed case 72
ck(f(('uint120',), D), T[i])  # fixed case 73
ck(f(('uint128',), D), T[i])  # fixed case 74
ck(f(('uint136',), D), T[i])  # fixed case 75
ck(f(('uint144',), D), T[i])  # fixed case 76
ck(f(('uint152',), D), T[i])  # fixed case 77
ck(f(('uint16',), D), T[i])  # fixed case 78
ck(f(('uint160',), D), T[i])  # fixed case 79
ck(f(('uint168',), D), T[i])  # fixed case 80
ck(f(('uint176',), D), T[i])  # fixed case 81
ck(f(('uint184',), D), T[i])  # fixed case 82
ck(f(('uint192',), D), T[i])  # fixed case 83
ck(f(('uint200',), D), T[i])  # fixed case 84
ck(f(('uint208',), D), T[i])  # fixed case 85
ck(f(('uint216',), D), T[i])  # fixed case 86
ck(f(('uint224',), D), T[i])  # fixed case 87
ck(f(('uint232',), D), T[i])  # fixed case 88
ck(f(('uint24',), D), T[i])  # fixed case 89
ck(f(('uint240',), D), T[i])  # fixed case 90
ck(f(('uint248',), D), T[i])  # fixed case 91
ck(f(('uint256',), D), T[i])  # fixed case 92
ck(f(('uint32',), D), T[i])  # fixed case 93
ck(f(('uint40',), D), T[i])  # fixed case 94
ck(f(('uint48',), D), T[i])  # fixed case 95
ck(f(('uint56',), D), T[i])  # fixed case 96
ck(f(('uint64',), D), T[i])  # fixed case 97
ck(f(('uint72',), D), T[i])  # fixed case 98
ck(f(('uint8',), D), T[i])  # fixed case 99
ck(f(('uint80',), D), T[i])  # fixed case 100
ck(f(('uint88',), D), T[i])  # fixed case 101
ck(f(('uint96',), D), T[i])  # fixed case 102
ck(f(('?',), D), T[A])  # fixed case 103



import pytest
import sys

from faster_eth_utils import (
    big_endian_to_int,
    decode_hex,
    int_to_big_endian,
    to_normalized_address,
)
from faster_eth_utils.toolz import (
    complement,
)
from hypothesis import (
    example,
    given,
    settings,
    strategies as st,
)

from faster_eth_abi import (
    decode,
    encode,
)
from faster_eth_abi.constants import (
    TT256M1,
)
from faster_eth_abi.decoding import (
    AddressDecoder,
    BooleanDecoder,
    BytesDecoder,
    ByteStringDecoder,
    ContextFramesBytesIO,
    DynamicArrayDecoder,
    SignedFixedDecoder,
    SignedIntegerDecoder,
    StringDecoder,
    TupleDecoder,
    UnsignedFixedDecoder,
    UnsignedIntegerDecoder,
)
from faster_eth_abi.encoding import (
    TextStringEncoder,
)
from faster_eth_abi.exceptions import (
    InsufficientDataBytes,
    NonEmptyPaddingBytes,
)
from faster_eth_abi.registry import (
    registry,
)
from faster_eth_abi.utils.numeric import (
    ceil32,
)
from faster_eth_abi.utils.padding import (
    zpad32,
)


def is_utf8_decodable(value):
    try:
        value.decode("utf-8")
    except UnicodeDecodeError:
        return False
    return True


def is_utf8_encodable(value):
    try:
        value.encode("utf-8")
    except UnicodeEncodeError:
        return False
    return True


def is_non_empty_non_null_byte_string(value):
    return value and big_endian_to_int(value) != 0


def is_valid_padding_bytes(padding_bytes, data_bytes):
    # Empty padding is always valid
    if len(padding_bytes) == 0:
        return True

    leading_data_bit_is_one = (data_bytes[0] & 0b10000000) == 0b10000000

    if leading_data_bit_is_one:
        # All padding bits must be 1
        if padding_bytes.replace(b"\xff", b"") == b"":
            return True
    else:
        # All padding bits must be 0
        if padding_bytes.replace(b"\x00", b"") == b"":
            return True

    return False


def all_bytes_equal(test_bytes, target):
    if sys.version_info.major < 3:
        return all(byte == chr(target) for byte in test_bytes)
    else:
        return all(byte == target for byte in test_bytes)


@settings(max_examples=250)
@given(
    integer_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_unsigned_int(integer_bit_size, stream_bytes, data_byte_size):
    if integer_bit_size % 8 != 0:
        with pytest.raises(ValueError):
            UnsignedIntegerDecoder(
                value_bit_size=integer_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif integer_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            UnsignedIntegerDecoder(
                value_bit_size=integer_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = UnsignedIntegerDecoder(
            value_bit_size=integer_bit_size,
            data_byte_size=data_byte_size,
        )

    stream = ContextFramesBytesIO(stream_bytes)
    actual_value = big_endian_to_int(stream_bytes[:data_byte_size])

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif actual_value > 2**integer_bit_size - 1:
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    integer_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
@example(8, b"\x00\x80", 2)
@example(8, b"\xff\xff", 2)
def test_decode_signed_int(integer_bit_size, stream_bytes, data_byte_size):
    if integer_bit_size % 8 != 0:
        with pytest.raises(ValueError):
            SignedIntegerDecoder(
                value_bit_size=integer_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    elif integer_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            SignedIntegerDecoder(
                value_bit_size=integer_bit_size,
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = SignedIntegerDecoder(
            value_bit_size=integer_bit_size,
            data_byte_size=data_byte_size,
        )

    stream = ContextFramesBytesIO(stream_bytes)

    padding_bytes = data_byte_size - integer_bit_size // 8

    raw_value = big_endian_to_int(stream_bytes[padding_bytes:data_byte_size])
    if raw_value >= 2 ** (integer_bit_size - 1):
        actual_value = raw_value - 2**integer_bit_size
    else:
        actual_value = raw_value

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif (
        actual_value >= 0 and not all_bytes_equal(stream_bytes[:padding_bytes], 0)
    ) or (actual_value < 0 and not all_bytes_equal(stream_bytes[:padding_bytes], 255)):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    _bytes=st.binary(min_size=0, max_size=256),
    pad_size=st.integers(min_value=0, max_value=32),
)
def test_decode_bytes(_bytes, pad_size):
    size_bytes = zpad32(int_to_big_endian(len(_bytes)))
    padded_bytes = _bytes + b"\x00" * pad_size
    stream_bytes = size_bytes + padded_bytes
    stream = ContextFramesBytesIO(stream_bytes)

    decoder = ByteStringDecoder()

    if len(padded_bytes) < ceil32(len(_bytes)):
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    decoded_value = decoder(stream)
    assert decoded_value == _bytes


@settings(max_examples=250)
@given(
    _strings=st.text(min_size=0, max_size=256).filter(is_utf8_encodable),
    pad_size=st.integers(min_value=0, max_value=32),
)
def test_decode_strings(_strings, pad_size):
    size_bytes = zpad32(int_to_big_endian(len(_strings.encode("utf-8"))))
    padded_bytes = _strings.encode("utf-8") + b"\x00" * pad_size
    stream_bytes = size_bytes + padded_bytes
    stream = ContextFramesBytesIO(stream_bytes)

    decoder = StringDecoder()

    if len(padded_bytes) < ceil32(len(_strings.encode("utf-8"))):
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    decoded_value = decoder(stream)
    assert decoded_value == _strings


@settings(max_examples=250)
@given(
    _bytes=st.binary(min_size=0, max_size=256).filter(complement(is_utf8_decodable)),
    pad_size=st.integers(min_value=0, max_value=32),
)
@pytest.mark.parametrize(
    "handle_string_errors,error",
    [
        ("strict", UnicodeError),
        ("ignore", None),
        ("replace", None),
        ("backslashreplace", None),
        ("surrogateescape", None),
    ],
)
def test_decode_strings_error_handling(_bytes, pad_size, handle_string_errors, error):
    size_bytes = zpad32(int_to_big_endian(len(_bytes)))
    padded_bytes = _bytes + b"\x00" * pad_size
    stream_bytes = size_bytes + padded_bytes
    stream = ContextFramesBytesIO(stream_bytes)
    decoder = StringDecoder(handle_string_errors=handle_string_errors)

    if len(padded_bytes) < ceil32(len(_bytes)):
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    if error:
        with pytest.raises(error):
            decoder(stream)
    else:
        assert decoder(stream) == _bytes.decode("utf-8", errors=handle_string_errors)


def test_string_decode_errors_are_strict_by_default():
    test_string = encode(["string"], ["cat"])
    bad_string = test_string[:65] + b"\xff" + test_string[66:]
    with pytest.raises(UnicodeError):
        decode(["string"], bad_string)


def test_registering_invalid_handle_string_errors_raises():
    test_string = encode(["string"], ["cat"])
    bad_string = test_string[:65] + b"\xff" + test_string[66:]
    registry.register(
        "nonexistent_handler",
        TextStringEncoder,
        StringDecoder(handle_string_errors="nonexistent"),
    )
    with pytest.raises(LookupError):
        decode(["nonexistent_handler"], bad_string)


@settings(max_examples=250)
@given(
    stream_bytes=st.binary(min_size=1, max_size=32),
    data_byte_size=st.integers(min_value=1, max_value=32),
)
def test_decode_boolean(stream_bytes, data_byte_size):
    stream = ContextFramesBytesIO(stream_bytes)

    decoder = BooleanDecoder(data_byte_size=data_byte_size)

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    padding_bytes = stream_bytes[:data_byte_size][:-1]
    if is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return

    byte_value = stream_bytes[data_byte_size - 1]

    if byte_value in {0, b"\x00"}:
        actual_value = False
    elif byte_value in {1, b"\x01"}:
        actual_value = True
    else:
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return

    decoded_value = decoder(stream)
    assert decoded_value is actual_value


@settings(max_examples=250)
@given(
    value_byte_size=st.integers(min_value=1, max_value=32),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_bytes_xx(value_byte_size, stream_bytes, data_byte_size):
    if value_byte_size > data_byte_size:
        with pytest.raises(ValueError):
            BytesDecoder(
                value_bit_size=value_byte_size * 8,
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = BytesDecoder(
            value_bit_size=value_byte_size * 8,
            data_byte_size=data_byte_size,
        )

    stream = ContextFramesBytesIO(stream_bytes)
    actual_value = stream_bytes[:value_byte_size]
    padding_bytes = stream_bytes[value_byte_size:data_byte_size]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    address_bytes=st.binary(min_size=0, max_size=32),
    padding_size=st.integers(min_value=10, max_value=14),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_address(address_bytes, padding_size, data_byte_size):
    stream_bytes = b"\x00" * padding_size + address_bytes
    if data_byte_size < 20:
        with pytest.raises(ValueError):
            AddressDecoder(
                data_byte_size=data_byte_size,
            )
        return
    else:
        decoder = AddressDecoder(
            data_byte_size=data_byte_size,
        )

    stream = ContextFramesBytesIO(stream_bytes)
    padding_bytes = stream_bytes[:data_byte_size][:-20]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return
    elif is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return
    else:
        decoded_value = decoder(stream)

    actual_value = to_normalized_address(stream_bytes[:data_byte_size][-20:])

    assert decoded_value == actual_value


@settings(max_examples=250)
@given(
    array_size=st.integers(min_value=0, max_value=32),
    array_values=st.lists(
        st.integers(min_value=0, max_value=TT256M1),
        min_size=0,
        max_size=64,
    ).map(tuple),
)
def test_decode_array_of_unsigned_integers(array_size, array_values):
    size_bytes = zpad32(int_to_big_endian(array_size))
    values_bytes = b"".join(zpad32(int_to_big_endian(v)) for v in array_values)
    stream_bytes = size_bytes + values_bytes

    decoder = DynamicArrayDecoder(
        item_decoder=UnsignedIntegerDecoder(value_bit_size=256),
    )
    stream = ContextFramesBytesIO(stream_bytes)

    if len(array_values) < array_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    actual_values = decoder(stream)
    assert actual_values == array_values[:array_size]


@pytest.mark.parametrize(
    "types,data,expected",
    (
        (
            ("address", "uint256"),
            (
                "0x"
                "000000000000000000000000abf7d8b5c1322b3e553d2fac90ff006c30f1b875"
                "0000000000000000000000000000000000000000000000000000005d21dba000"
            ),
            ("0xabf7d8b5c1322b3e553d2fac90ff006c30f1b875", 400000000000),
        ),
        (
            ("uint256", "bytes"),
            (
                "0x"
                "0000000000000000000000000000000000000000000000000000000000000000"
                "0000000000000000000000000000000000000000000000000000000000000040"
                "0000000000000000000000000000000000000000000000000000000000000000"
                "0000000000000000000000000000000000000000000000000000000000000000"
            ),
            (0, b""),
        ),
    ),
)
def test_tuple_decoder(types, data, expected):
    decoders = [registry.get_decoder(t) for t in types]
    decoder = TupleDecoder(decoders=decoders)
    stream = ContextFramesBytesIO(decode_hex(data))
    actual = decoder(stream)
    assert actual == expected


@settings(max_examples=250)
@given(
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    frac_places=st.integers(min_value=1, max_value=80),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_unsigned_fixed(
    value_bit_size, frac_places, stream_bytes, data_byte_size
):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            UnsignedFixedDecoder(
                value_bit_size=value_bit_size,
                frac_places=frac_places,
                data_byte_size=data_byte_size,
            )
        return

    decoder = UnsignedFixedDecoder(
        value_bit_size=value_bit_size,
        frac_places=frac_places,
        data_byte_size=data_byte_size,
    )

    stream = ContextFramesBytesIO(stream_bytes)
    padding_bytes = stream_bytes[:data_byte_size][
        : data_byte_size - value_bit_size // 8
    ]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    if is_non_empty_non_null_byte_string(padding_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return

    # Ensure no exceptions
    decoder(stream)


@settings(max_examples=250)
@given(
    value_bit_size=st.integers(min_value=1, max_value=32).map(lambda v: v * 8),
    frac_places=st.integers(min_value=1, max_value=80),
    stream_bytes=st.binary(min_size=0, max_size=32),
    data_byte_size=st.integers(min_value=0, max_value=32),
)
def test_decode_signed_fixed(value_bit_size, frac_places, stream_bytes, data_byte_size):
    if value_bit_size > data_byte_size * 8:
        with pytest.raises(ValueError):
            SignedFixedDecoder(
                value_bit_size=value_bit_size,
                frac_places=frac_places,
                data_byte_size=data_byte_size,
            )
        return

    decoder = SignedFixedDecoder(
        value_bit_size=value_bit_size,
        frac_places=frac_places,
        data_byte_size=data_byte_size,
    )

    stream = ContextFramesBytesIO(stream_bytes)

    padding_offset = data_byte_size - value_bit_size // 8
    data_offset = padding_offset + value_bit_size // 8

    padding_bytes = stream_bytes[:data_byte_size][:padding_offset]
    data_bytes = stream_bytes[:data_byte_size][padding_offset:data_offset]

    if len(stream_bytes) < data_byte_size:
        with pytest.raises(InsufficientDataBytes):
            decoder(stream)
        return

    if not is_valid_padding_bytes(padding_bytes, data_bytes):
        with pytest.raises(NonEmptyPaddingBytes):
            decoder(stream)
        return

    actual_value = decoder(stream)

    if padding_bytes:
        if actual_value >= 0:
            assert bytes(set(padding_bytes)) == b"\x00"
        else:
            assert bytes(set(padding_bytes)) == b"\xff"



from faster_eth_abi import (
    decode,
    encode,
)
from faster_eth_abi.decoding import (
    BaseDecoder,
)
from faster_eth_abi.encoding import (
    BaseEncoder,
)
from faster_eth_abi.exceptions import (
    DecodingError,
    EncodingError,
)
from faster_eth_abi.registry import (
    registry,
)

NULL_ENCODING = b"\x00" * 32


def encode_null(x):
    if x is not None:
        raise EncodingError("Unsupported value")

    return NULL_ENCODING


def decode_null(stream):
    if stream.read(32) != NULL_ENCODING:
        raise DecodingError("Not enough data or wrong data")

    return None


class EncodeNull(BaseEncoder):
    word_width = None

    @classmethod
    def from_type_str(cls, type_str, registry):
        word_width = int(type_str[4:])
        return cls(word_width=word_width)

    def encode(self, value):
        self.validate_value(value)
        return NULL_ENCODING * self.word_width

    def validate_value(self, value):
        if value is not None:
            raise EncodingError("Unsupported value")


class DecodeNull(BaseDecoder):
    word_width = None

    @classmethod
    def from_type_str(cls, type_str, registry):
        word_width = int(type_str[4:])
        return cls(word_width=word_width)

    def decode(self, stream):
        byts = stream.read(32 * self.word_width)
        if byts != NULL_ENCODING * self.word_width:
            raise DecodingError("Not enough data or wrong data")

        return None


def test_register_and_use_callables():
    registry.register("null", encode_null, decode_null)

    try:
        assert encode(["null"], [None]) == NULL_ENCODING

        (decoded_null_data,) = decode(["null"], NULL_ENCODING)
        assert decoded_null_data is None

        encoded_tuple = encode(["(int,null)"], [(1, None)])
        assert encoded_tuple == b"\x00" * 31 + b"\x01" + NULL_ENCODING

        (decoded_tuple,) = decode(["(int,null)"], encoded_tuple)
        assert decoded_tuple == (1, None)
    finally:
        registry.unregister("null")


def test_register_and_use_coder_classes():
    registry.register(
        lambda x: x.startswith("null"),
        EncodeNull,
        DecodeNull,
        label="null",
    )

    try:
        assert encode(["null2"], [None]) == NULL_ENCODING * 2

        (decoded_null_data,) = decode(["null2"], NULL_ENCODING * 2)
        assert decoded_null_data is None

        encoded_tuple = encode(["(int,null2)"], [(1, None)])
        assert encoded_tuple == b"\x00" * 31 + b"\x01" + NULL_ENCODING * 2

        (decoded_tuple,) = decode(["(int,null2)"], encoded_tuple)
        assert decoded_tuple == (1, None)
    finally:
        registry.unregister("null")



#!/usr/bin/env python
import sys
from typing import (
    List,
)

from mypyc.build import (
    mypycify,
)
from setuptools import (
    find_packages,
    setup,
)

REQUIREMENTS = []
HYPOTHESIS_REQUIREMENT = "hypothesis>=6.22.0,<6.108.7"


def parse_requirements(filename: str) -> List[str]:
    lines = []
    with open(filename) as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            if line.startswith("-r"):
                lines.extend(
                    line
                    for line in parse_requirements(line[2:].strip())
                    if line not in REQUIREMENTS
                )
            else:
                lines.append(line.strip())
    return lines


REQUIREMENTS.extend(parse_requirements("requirements.txt"))

extras_require = {
    "dev": parse_requirements("requirements-dev.txt"),
    "docs": [
        "sphinx>=6.0.0",
        "sphinx-autobuild>=2021.3.14",
        "sphinx_rtd_theme>=1.0.0",
        "towncrier>=24,<26",
    ],
    "test": [
        "pytest>=7.0.0",
        "pytest-timeout>=2.0.0",
        "pytest-xdist>=2.4.0",
        "pytest-pythonpath>=0.7.1",
        "eth-hash[pycryptodome]",
        HYPOTHESIS_REQUIREMENT,
    ],
    "tools": [
        HYPOTHESIS_REQUIREMENT,
    ],
    "codspeed": [
        "pytest>=7.0.0",
        "pytest-codspeed>=4.2,<4.3",
        "pytest-test-groups",
    ],
    "mypy": parse_requirements("requirements-mypy.txt"),
}

extras_require["dev"] = (
    extras_require["dev"] + extras_require["docs"] + extras_require["test"]
)


with open("./README.md") as readme:
    long_description = readme.read()


skip_mypyc = any(
    cmd in sys.argv
    for cmd in ("sdist", "egg_info", "--name", "--version", "--help", "--help-commands")
)

if skip_mypyc:
    ext_modules = []
else:
    mypycify_kwargs = {"strict_dunder_typing": True}
    if sys.version_info >= (3, 9):
        mypycify_kwargs["group_name"] = "faster_eth_abi"

    ext_modules = mypycify(
        [
            "faster_eth_abi/_codec.py",
            "faster_eth_abi/_decoding.py",
            "faster_eth_abi/_encoding.py",
            "faster_eth_abi/_grammar.py",
            "faster_eth_abi/abi.py",
            "faster_eth_abi/constants.py",
            # "faster_eth_abi/exceptions.py",  segfaults on mypyc 1.18.2
            "faster_eth_abi/from_type_str.py",
            # "faster_eth_abi/io.py",
            "faster_eth_abi/packed.py",
            "faster_eth_abi/tools",
            "faster_eth_abi/utils",
            "--pretty",
            "--install-types",
            # all of these are safe to disable long term
            "--disable-error-code=override",
            "--disable-error-code=unused-ignore",
            "--disable-error-code=no-any-return",
        ],
        **mypycify_kwargs,
    )


setup(
    name="faster_eth_abi",
    # *IMPORTANT*: Don't manually change the version here. See Contributing docs for the release process.
    version="5.2.16",
    description="""A ~2-6x faster fork of eth_abi: Python utilities for working with Ethereum ABI definitions, especially encoding and decoding. Implemented in C.""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="The Ethereum Foundation",
    author_email="snakecharmers@ethereum.org",
    url="https://github.com/BobTheBuidler/faster-eth-abi",
    project_urls={
        "Documentation": "https://eth-abi.readthedocs.io/en/stable/",
        "Release Notes": "https://github.com/BobTheBuidler/faster-eth-abi/releases",
        "Issues": "https://github.com/BobTheBuidler/faster-eth-abi/issues",
        "Source - Precompiled (.py)": "https://github.com/BobTheBuidler/faster-eth-utils/tree/master/faster_eth_utils",
        "Source - Compiled (.c)": "https://github.com/BobTheBuidler/faster-eth-utils/tree/master/build",
        "Benchmarks": "https://github.com/BobTheBuidler/faster-eth-utils/tree/master/benchmarks",
        "Benchmarks - Results": "https://github.com/BobTheBuidler/faster-eth-utils/tree/master/benchmarks/results",
        "Original": "https://github.com/ethereum/eth-abi",
    },
    include_package_data=True,
    install_requires=REQUIREMENTS,
    python_requires=">=3.8, <4",
    extras_require=extras_require,
    license="MIT",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages(
        exclude=[
            "benchmarks",
            "benchmarks.*",
            "scripts",
            "scripts.*",
            "tests",
            "tests.*",
            "eth-abi-stubs",
            "eth-abi-stubs.*",
        ]
    ),
    ext_modules=ext_modules,
    package_data={"faster_eth_abi": ["py.typed"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
)



"""Private helpers for ABI type string grammar and parsing, intended for C compilation.

This file exists because the original grammar.py is not ready to be fully compiled to C.
This module contains functions and logic that we do wish to compile.
"""
import re
from typing import (
    Any,
    Final,
    Generic,
    Literal,
    NewType,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    final,
)

from eth_typing.abi import (
    TypeStr,
)
from mypy_extensions import (
    mypyc_attr,
)
from parsimonious.nodes import (
    Node,
)
from typing_extensions import (
    Self,
)

from faster_eth_abi.exceptions import (
    ABITypeError,
)

TYPE_ALIASES: Final = {
    "int": "int256",
    "uint": "uint256",
    "fixed": "fixed128x18",
    "ufixed": "ufixed128x18",
    "function": "bytes24",
    "byte": "bytes1",
}

TYPE_ALIAS_RE: Final = re.compile(
    rf"\b({'|'.join(map(re.escape, TYPE_ALIASES.keys()))})\b"
)


IntSubtype = NewType("IntSubtype", int)
FixedSubtype = NewType("FixedSubtype", Tuple[int, int])
Subtype = Union[IntSubtype, FixedSubtype]
TSub = TypeVar("TSub", IntSubtype, FixedSubtype, Literal[None])


@mypyc_attr(allow_interpreted_subclasses=True)
class ABIType:
    """
    Base class for results of type string parsing operations.
    """

    __slots__ = ("arrlist", "node")

    def __init__(
        self, arrlist: Optional[Sequence[str]] = None, node: Optional[Node] = None
    ) -> None:
        self.arrlist: Final = arrlist
        """
        The list of array dimensions for a parsed type.  Equal to ``None`` if
        type string has no array dimensions.
        """

        self.node: Final = node
        """
        The parsimonious ``Node`` instance associated with this parsed type.
        Used to generate error messages for invalid types.
        """

    def __repr__(self) -> str:  # pragma: no cover
        return f"<{type(self).__qualname__} {self.to_type_str()!r}>"

    def __eq__(self, other: Any) -> bool:
        # Two ABI types are equal if their string representations are equal
        return type(self) is type(other) and self.to_type_str() == other.to_type_str()

    def to_type_str(self) -> TypeStr:  # pragma: no cover
        """
        Returns the string representation of an ABI type.  This will be equal to
        the type string from which it was created.
        """
        raise NotImplementedError("Must implement `to_type_str`")

    @property
    def item_type(self) -> Self:
        """
        If this type is an array type, equal to an appropriate
        :class:`~faster_eth_abi.grammar.ABIType` instance for the array's items.
        """
        raise NotImplementedError("Must implement `item_type`")

    def validate(self) -> None:  # pragma: no cover
        """
        Validates the properties of an ABI type against the solidity ABI spec:

        https://solidity.readthedocs.io/en/develop/abi-spec.html

        Raises :class:`~faster_eth_abi.exceptions.ABITypeError` if validation fails.
        """
        raise NotImplementedError("Must implement `validate`")

    @final
    def invalidate(self, error_msg: str) -> NoReturn:
        # Invalidates an ABI type with the given error message.  Expects that a
        # parsimonious node was provided from the original parsing operation
        # that yielded this type.
        node = self.node

        raise ABITypeError(
            f"For '{node.text}' type at column {node.start + 1} "
            f"in '{node.full_text}': {error_msg}"
        )

    @final
    @property
    def is_array(self) -> bool:
        """
        Equal to ``True`` if a type is an array type (i.e. if it has an array
        dimension list).  Otherwise, equal to ``False``.
        """
        return self.arrlist is not None

    @property
    def is_dynamic(self) -> bool:
        """
        Equal to ``True`` if a type has a dynamically sized encoding.
        Otherwise, equal to ``False``.
        """
        raise NotImplementedError("Must implement `is_dynamic`")

    @final
    @property
    def _has_dynamic_arrlist(self) -> bool:
        return self.is_array and any(len(dim) == 0 for dim in self.arrlist)


TComp = TypeVar("TComp", bound=ABIType)


@mypyc_attr(allow_interpreted_subclasses=True)
class TupleType(ABIType):
    """
    Represents the result of parsing a tuple type string e.g. "(int,bool)".
    """

    __slots__ = ("components",)

    def __init__(
        self,
        components: Tuple[TComp, ...],
        arrlist: Optional[Sequence[str]] = None,
        *,
        node: Optional[Node] = None,
    ) -> None:
        super().__init__(arrlist, node)

        self.components: Final = components
        """
        A tuple of :class:`~faster_eth_abi.grammar.ABIType` instances for each of the
        tuple type's components.
        """

    def to_type_str(self) -> TypeStr:
        arrlist = self.arrlist

        if isinstance(arrlist, tuple):
            arrlist = "".join(map(repr, map(list, arrlist)))
        else:
            arrlist = ""

        return f"({','.join(c.to_type_str() for c in self.components)}){arrlist}"

    @property
    def item_type(self) -> Self:
        if not self.is_array:
            raise ValueError(
                f"Cannot determine item type for non-array type '{self.to_type_str()}'"
            )

        arrlist = self.arrlist[:-1] or None  # type: ignore [index]
        cls = type(self)
        if cls is TupleType:
            return TupleType(self.components, arrlist, node=self.node)  # type: ignore [return-value]
        else:
            return cls(self.components, arrlist, node=self.node)

    def validate(self) -> None:
        for c in self.components:
            c.validate()

    @property
    def is_dynamic(self) -> bool:
        if self._has_dynamic_arrlist:
            return True

        return any(c.is_dynamic for c in self.components)


@mypyc_attr(allow_interpreted_subclasses=True)
class BasicType(ABIType, Generic[TSub]):
    """
    Represents the result of parsing a basic type string e.g. "uint", "address",
    "ufixed128x19[][2]".
    """

    __slots__ = ("base", "sub")

    def __init__(
        self,
        base: str,
        sub: Optional[TSub] = None,
        arrlist: Optional[Sequence] = None,
        *,
        node: Optional[Node] = None,
    ) -> None:
        super().__init__(arrlist, node)

        self.base: Final = base
        """The base of a basic type e.g. "uint" for "uint256" etc."""

        self.sub: Final = sub
        """
        The sub type of a basic type e.g. ``256`` for "uint256" or ``(128, 18)``
        for "ufixed128x18" etc.  Equal to ``None`` if type string has no sub
        type.
        """

    def to_type_str(self) -> TypeStr:
        sub, arrlist = self.sub, self.arrlist

        if isinstance(sub, int):
            substr = str(sub)
        elif isinstance(sub, tuple):
            substr = "x".join(map(str, sub))
        else:
            substr = ""

        if isinstance(arrlist, tuple):
            return self.base + substr + "".join(map(repr, map(list, arrlist)))
        else:
            return self.base + substr

    @property
    def item_type(self) -> Self:
        if not self.is_array:
            raise ValueError(
                f"Cannot determine item type for non-array type '{self.to_type_str()}'"
            )

        cls = type(self)
        arrlist = self.arrlist[:-1] or None  # type: ignore [index]
        if cls is BasicType:
            return BasicType(self.base, self.sub, arrlist, node=self.node)  # type: ignore [return-value]
        else:
            return cls(self.base, self.sub, arrlist, node=self.node)

    @property
    def is_dynamic(self) -> bool:
        if self._has_dynamic_arrlist:
            return True

        base = self.base
        if base == "string":
            return True

        if base == "bytes" and self.sub is None:
            return True

        return False

    def validate(self) -> None:
        base, sub = self.base, self.sub

        # Check validity of string type
        if base == "string":
            if sub is not None:
                self.invalidate("string type cannot have suffix")

        # Check validity of bytes type
        elif base == "bytes":
            if not (sub is None or isinstance(sub, int)):
                self.invalidate(
                    "bytes type must have either no suffix or a numerical suffix"
                )

            if isinstance(sub, int) and sub > 32:
                self.invalidate("maximum 32 bytes for fixed-length bytes")

        # Check validity of integer type
        elif base in ("int", "uint"):
            if not isinstance(sub, int):
                self.invalidate("integer type must have numerical suffix")

            if sub < 8 or sub > 256:
                self.invalidate("integer size out of bounds (max 256 bits)")

            if sub % 8 != 0:
                self.invalidate("integer size must be multiple of 8")

        # Check validity of fixed type
        elif base in ("fixed", "ufixed"):
            if not isinstance(sub, tuple):
                self.invalidate(
                    "fixed type must have suffix of form <bits>x<exponent>, "
                    "e.g. 128x19",
                )

            bits, minus_e = sub

            if bits < 8 or bits > 256:
                self.invalidate("fixed size out of bounds (max 256 bits)")

            if bits % 8 != 0:
                self.invalidate("fixed size must be multiple of 8")

            if minus_e < 1 or minus_e > 80:
                self.invalidate(
                    f"fixed exponent size out of bounds, {minus_e} must be in 1-80"
                )

        # Check validity of hash type
        elif base == "hash":
            if not isinstance(sub, int):
                self.invalidate("hash type must have numerical suffix")

        # Check validity of address type
        elif base == "address":
            if sub is not None:
                self.invalidate("address cannot have suffix")


BytesType = BasicType[IntSubtype]
FixedType = BasicType[FixedSubtype]


def normalize(type_str: TypeStr) -> TypeStr:
    """
    Normalizes a type string into its canonical version e.g. the type string
    'int' becomes 'int256', etc.

    :param type_str: The type string to be normalized.
    :returns: The canonical version of the input type string.
    """
    return TYPE_ALIAS_RE.sub(__normalize, type_str)


def __normalize(match: "re.Match[str]") -> str:
    return TYPE_ALIASES[match.group(0)]



import abc
from . import (
    decoding as decoding,
    encoding as encoding,
    exceptions as exceptions,
    grammar as grammar,
)
from .base import BaseCoder as BaseCoder
from .exceptions import (
    MultipleEntriesFound as MultipleEntriesFound,
    NoEntriesFound as NoEntriesFound,
)
from .io import ContextFramesBytesIO as ContextFramesBytesIO
from _typeshed import Incomplete
from eth_typing import TypeStr
from typing import Any, Callable, Final, Generic, Iterator, TypeVar
from typing_extensions import ParamSpec, Self

T = TypeVar("T")
P = ParamSpec("P")
Lookup = TypeStr | Callable[[TypeStr], bool]
EncoderCallable = Callable[[Any], bytes]
DecoderCallable = Callable[[ContextFramesBytesIO], Any]
Encoder = EncoderCallable | type[encoding.BaseEncoder]
Decoder = DecoderCallable | type[decoding.BaseDecoder]

class Copyable(abc.ABC, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def copy(self): ...
    def __copy__(self): ...
    def __deepcopy__(self, *args): ...

class PredicateMapping(Copyable, Generic[T]):
    """
    Acts as a mapping from predicate functions to values.  Values are retrieved
    when their corresponding predicate matches a given input.  Predicates can
    also be labeled to facilitate removal from the mapping.
    """

    def __init__(self, name: str) -> None: ...
    def add(self, predicate: Lookup, value: T, label: str | None = None) -> None: ...
    def find(self, type_str: TypeStr) -> T: ...
    def remove_by_equality(self, predicate: Lookup) -> None: ...
    def remove_by_label(self, label: str) -> None: ...
    def remove(self, predicate_or_label: Lookup | str) -> None: ...
    def copy(self) -> Self: ...

class Predicate(Generic[T]):
    """
    Represents a predicate function to be used for type matching in
    ``ABIRegistry``.
    """

    def __init__(self) -> None: ...
    def __call__(self, arg: TypeStr) -> None: ...
    def __iter__(self) -> Iterator[T]: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: Predicate) -> bool: ...

class Equals(Predicate[str]):
    """
    A predicate that matches any input equal to `value`.
    """

    value: Final[Incomplete]
    def __init__(self, value: str) -> None: ...
    def __call__(self, other: TypeStr) -> bool: ...
    def __iter__(self) -> Iterator[str]: ...

class BaseEquals(Predicate[str | bool | None]):
    """
    A predicate that matches a basic type string with a base component equal to
    `value` and no array component.  If `with_sub` is `True`, the type string
    must have a sub component to match.  If `with_sub` is `False`, the type
    string must *not* have a sub component to match.  If `with_sub` is None,
    the type string's sub component is ignored.
    """

    base: Final[Incomplete]
    with_sub: Final[Incomplete]
    def __init__(self, base: TypeStr, *, with_sub: bool | None = None) -> None: ...
    def __call__(self, type_str: TypeStr) -> bool: ...
    def __iter__(self) -> Iterator[str | bool | None]: ...

def has_arrlist(type_str: TypeStr) -> bool:
    """
    A predicate that matches a type string with an array dimension list.
    """

def is_base_tuple(type_str: TypeStr) -> bool:
    """
    A predicate that matches a tuple type with no array dimension list.
    """

class BaseRegistry: ...

class ABIRegistry(Copyable, BaseRegistry):
    get_encoder: Incomplete
    get_decoder: Incomplete
    get_tuple_encoder: Incomplete
    get_tuple_decoder: Incomplete
    def __init__(self) -> None: ...
    @_clear_encoder_cache
    def register_encoder(
        self, lookup: Lookup, encoder: Encoder, label: str | None = None
    ) -> None:
        """
        Registers the given ``encoder`` under the given ``lookup``.  A unique
        string label may be optionally provided that can be used to refer to
        the registration by name.  For more information about arguments, refer
        to :any:`register`.
        """

    @_clear_encoder_cache
    def unregister_encoder(self, lookup_or_label: Lookup) -> None:
        """
        Unregisters an encoder in the registry with the given lookup or label.
        If ``lookup_or_label`` is a string, the encoder with the label
        ``lookup_or_label`` will be unregistered.  If it is an function, the
        encoder with the lookup function ``lookup_or_label`` will be
        unregistered.
        """

    @_clear_decoder_cache
    def register_decoder(
        self, lookup: Lookup, decoder: Decoder, label: str | None = None
    ) -> None:
        """
        Registers the given ``decoder`` under the given ``lookup``.  A unique
        string label may be optionally provided that can be used to refer to
        the registration by name.  For more information about arguments, refer
        to :any:`register`.
        """

    @_clear_decoder_cache
    def unregister_decoder(self, lookup_or_label: Lookup) -> None:
        """
        Unregisters a decoder in the registry with the given lookup or label.
        If ``lookup_or_label`` is a string, the decoder with the label
        ``lookup_or_label`` will be unregistered.  If it is an function, the
        decoder with the lookup function ``lookup_or_label`` will be
        unregistered.
        """

    def register(
        self,
        lookup: Lookup,
        encoder: Encoder,
        decoder: Decoder,
        label: str | None = None,
    ) -> None:
        """
        Registers the given ``encoder`` and ``decoder`` under the given
        ``lookup``.  A unique string label may be optionally provided that can
        be used to refer to the registration by name.

        :param lookup: A type string or type string matcher function
            (predicate).  When the registry is queried with a type string
            ``query`` to determine which encoder or decoder to use, ``query``
            will be checked against every registration in the registry.  If a
            registration was created with a type string for ``lookup``, it will
            be considered a match if ``lookup == query``.  If a registration
            was created with a matcher function for ``lookup``, it will be
            considered a match if ``lookup(query) is True``.  If more than one
            registration is found to be a match, then an exception is raised.

        :param encoder: An encoder callable or class to use if ``lookup``
            matches a query.  If ``encoder`` is a callable, it must accept a
            python value and return a ``bytes`` value.  If ``encoder`` is a
            class, it must be a valid subclass of :any:`encoding.BaseEncoder`
            and must also implement the :any:`from_type_str` method on
            :any:`base.BaseCoder`.

        :param decoder: A decoder callable or class to use if ``lookup``
            matches a query.  If ``decoder`` is a callable, it must accept a
            stream-like object of bytes and return a python value.  If
            ``decoder`` is a class, it must be a valid subclass of
            :any:`decoding.BaseDecoder` and must also implement the
            :any:`from_type_str` method on :any:`base.BaseCoder`.

        :param label: An optional label that can be used to refer to this
            registration by name.  This label can be used to unregister an
            entry in the registry via the :any:`unregister` method and its
            variants.
        """

    def unregister(self, label: str | None) -> None:
        """
        Unregisters the entries in the encoder and decoder registries which
        have the label ``label``.
        """

    def has_encoder(self, type_str: TypeStr) -> bool:
        """
        Returns ``True`` if an encoder is found for the given type string
        ``type_str``.  Otherwise, returns ``False``.  Raises
        :class:`~faster_eth_abi.exceptions.MultipleEntriesFound` if multiple encoders
        are found.
        """

    def copy(self) -> Self:
        """
        Copies a registry such that new registrations can be made or existing
        registrations can be unregistered without affecting any instance from
        which a copy was obtained.  This is useful if an existing registry
        fulfills most of a user's needs but requires one or two modifications.
        In that case, a copy of that registry can be obtained and the necessary
        changes made without affecting the original registry.
        """

registry: Incomplete
is_int: Incomplete
is_int8: Incomplete
is_int16: Incomplete
is_uint: Incomplete
is_uint8: Incomplete
is_uint16: Incomplete
registry_packed: Incomplete



import pytest

from faster_eth_abi.decoding import (
    ContextFramesBytesIO,
)


@pytest.fixture
def byte_string():
    return bytes(range(100))


@pytest.fixture
def byte_stream(byte_string):
    return ContextFramesBytesIO(byte_string)


def test_read_seek_tell_work_as_normal(byte_string, byte_stream):
    assert byte_stream.tell() == 0
    assert byte_stream.read() == byte_string
    assert byte_stream.tell() == len(byte_string)

    byte_stream.seek(0)

    assert byte_stream.tell() == 0
    assert byte_stream.read() == byte_string
    assert byte_stream.tell() == len(byte_string)


def test_pushing_and_popping_frames(byte_string, byte_stream):
    byte_stream.push_frame(5)
    assert byte_stream.read(5) == byte_string[5:10]
    byte_stream.seek_in_frame(1)
    assert byte_stream.read(1) == byte_string[6:7]

    byte_stream.seek_in_frame(5)
    byte_stream.push_frame(2)
    assert byte_stream.read(5) == byte_string[7:12]
    byte_stream.seek_in_frame(2)
    assert byte_stream.read(1) == byte_string[9:10]

    byte_stream.pop_frame()
    assert byte_stream.read(5) == byte_string[10:15]

    byte_stream.pop_frame()
    assert byte_stream.read(5) == byte_string[0:5]

    with pytest.raises(IndexError):
        byte_stream.pop_frame()



import eth_abi.exceptions

class EncodingError(eth_abi.exceptions.EncodingError):
    """
    Base exception for any error that occurs during encoding.
    """

class EncodingTypeError(EncodingError, eth_abi.exceptions.EncodingTypeError):
    """
    Raised when trying to encode a python value whose type is not supported for
    the output ABI type.
    """

class IllegalValue(EncodingError, eth_abi.exceptions.IllegalValue):
    """
    Raised when trying to encode a python value with the correct type but with
    a value that is not considered legal for the output ABI type.

    .. code-block:: python

        fixed128x19_encoder(Decimal('NaN'))  # cannot encode NaN

    """

class ValueOutOfBounds(IllegalValue, eth_abi.exceptions.ValueOutOfBounds):
    """
    Raised when trying to encode a python value with the correct type but with
    a value that appears outside the range of valid values for the output ABI
    type.

    .. code-block:: python

        ufixed8x1_encoder(Decimal('25.6'))  # out of bounds

    """

class DecodingError(eth_abi.exceptions.DecodingError):
    """
    Base exception for any error that occurs during decoding.
    """

class InsufficientDataBytes(DecodingError, eth_abi.exceptions.InsufficientDataBytes):
    """
    Raised when there are insufficient data to decode a value for a given ABI type.
    """

class NonEmptyPaddingBytes(DecodingError, eth_abi.exceptions.NonEmptyPaddingBytes):
    """
    Raised when the padding bytes of an ABI value are malformed.
    """

class InvalidPointer(DecodingError, eth_abi.exceptions.InvalidPointer):
    """
    Raised when the pointer to a value in the ABI encoding is invalid.
    """

class ParseError(eth_abi.exceptions.ParseError):
    """
    Raised when an ABI type string cannot be parsed.
    """

class ABITypeError(eth_abi.exceptions.ABITypeError):
    """
    Raised when a parsed ABI type has inconsistent properties; for example,
    when trying to parse the type string ``'uint7'`` (which has a bit-width
    that is not congruent with zero modulo eight).
    """

class PredicateMappingError(eth_abi.exceptions.PredicateMappingError):
    """
    Raised when an error occurs in a registry's internal mapping.
    """

class NoEntriesFound(PredicateMappingError, eth_abi.exceptions.NoEntriesFound):
    """
    Raised when no registration is found for a type string in a registry's
    internal mapping.

    .. warning::

        In a future version of ``faster-eth-abi``, this error class will no longer
        inherit from ``ValueError``.
    """

class MultipleEntriesFound(
    PredicateMappingError, eth_abi.exceptions.MultipleEntriesFound
):
    """
    Raised when multiple registrations are found for a type string in a
    registry's internal mapping.  This error is non-recoverable and indicates
    that a registry was configured incorrectly.  Registrations are expected to
    cover completely distinct ranges of type strings.

    .. warning::

        In a future version of ``faster-eth-abi``, this error class will no longer
        inherit from ``ValueError``.
    """



import pytest
import copy
import itertools

from faster_eth_abi import (
    decoding,
    encoding,
    exceptions,
)
from faster_eth_abi.registry import (
    ABIRegistry,
    BaseEquals,
    has_arrlist,
    registry as default_registry,
)


@pytest.fixture
def registry():
    reg = ABIRegistry()

    reg.register(
        has_arrlist,
        encoding.BaseArrayEncoder,
        decoding.BaseArrayDecoder,
        label="has_arrlist",
    )
    reg.register(
        BaseEquals("address"),
        encoding.AddressEncoder,
        decoding.AddressDecoder,
        label="address",
    )

    return reg


@pytest.mark.parametrize(
    "type_str, encoder_class, decoder_class",
    (
        ("address", encoding.AddressEncoder, decoding.AddressDecoder),
        ("bool", encoding.BooleanEncoder, decoding.BooleanDecoder),
        ("bytes12", encoding.BytesEncoder, decoding.BytesDecoder),
        ("function", encoding.BytesEncoder, decoding.BytesDecoder),
        ("bytes", encoding.ByteStringEncoder, decoding.ByteStringDecoder),
        ("int", encoding.SignedIntegerEncoder, decoding.SignedIntegerDecoder),
        ("int128", encoding.SignedIntegerEncoder, decoding.SignedIntegerDecoder),
        ("string", encoding.TextStringEncoder, decoding.StringDecoder),
        ("uint", encoding.UnsignedIntegerEncoder, decoding.UnsignedIntegerDecoder),
        ("uint8", encoding.UnsignedIntegerEncoder, decoding.UnsignedIntegerDecoder),
        ("int[]", encoding.DynamicArrayEncoder, decoding.DynamicArrayDecoder),
        ("int[2]", encoding.SizedArrayEncoder, decoding.SizedArrayDecoder),
        ("int[2][]", encoding.DynamicArrayEncoder, decoding.DynamicArrayDecoder),
        ("int[][2]", encoding.SizedArrayEncoder, decoding.SizedArrayDecoder),
    ),
)
def test_default_registry_gives_correct_basic_coders(
    type_str, encoder_class, decoder_class
):
    assert isinstance(default_registry.get_encoder(type_str), encoder_class)
    assert isinstance(default_registry.get_decoder(type_str), decoder_class)


def test_cache_resets_after_register_and_register_works(registry: ABIRegistry):
    # Populate cache
    registry.get_encoder("address")
    registry.get_decoder("address")

    # Perform cache resetting action
    registry.register(
        BaseEquals("address", with_sub=False),
        encoding.AddressEncoder,
        decoding.AddressDecoder,
        label="address without sub",
    )

    # Confirm cache reset
    with pytest.raises(exceptions.MultipleEntriesFound):
        registry.get_encoder("address")
    with pytest.raises(exceptions.MultipleEntriesFound):
        registry.get_decoder("address")


def test_cache_resets_after_unregister_and_unregister_works(registry: ABIRegistry):
    # Populate cache
    registry.get_encoder("address")
    registry.get_decoder("address")

    # Perform cache resetting action
    registry.unregister("address")

    # Confirm cache reset
    with pytest.raises(exceptions.NoEntriesFound):
        registry.get_encoder("address")
    with pytest.raises(exceptions.NoEntriesFound):
        registry.get_decoder("address")


def test_can_register_and_unregister_string_lookups(registry: ABIRegistry):
    registry.register(
        "bool",
        encoding.BooleanEncoder,
        decoding.BooleanDecoder,
    )

    assert isinstance(registry.get_encoder("bool"), encoding.BooleanEncoder)
    assert isinstance(registry.get_decoder("bool"), decoding.BooleanDecoder)

    registry.unregister("bool")

    with pytest.raises(exceptions.NoEntriesFound):
        registry.get_encoder("bool")
    with pytest.raises(exceptions.NoEntriesFound):
        registry.get_decoder("bool")


def test_registry_should_reject_unknown_types(registry: ABIRegistry):
    with pytest.raises(TypeError):
        registry.register(True, None, None)
    with pytest.raises(TypeError):
        registry.unregister(True)


def test_can_unregister_by_equality(registry: ABIRegistry):
    registry.unregister(BaseEquals("address"))

    with pytest.raises(exceptions.NoEntriesFound):
        registry.get_encoder("address")
    with pytest.raises(exceptions.NoEntriesFound):
        registry.get_decoder("address")


def test_can_register_simple_callables(registry: ABIRegistry):
    def encode_bool(x):
        return x

    decode_bool = encode_bool

    registry.register("bool", encode_bool, decode_bool)

    assert registry.get_encoder("bool") is encode_bool
    assert registry.get_decoder("bool") is decode_bool


def test_unregister_unknown_lookups(registry: ABIRegistry):
    with pytest.raises(KeyError, match=r"Matcher .* not found in encoder registry"):
        registry.unregister(lambda x: x)
    with pytest.raises(KeyError, match=r"Label .* not found in encoder registry"):
        registry.unregister("foo")


def test_looking_up_unparsable_type_causes_error(registry: ABIRegistry):
    with pytest.raises(exceptions.ParseError, match="Parse error at"):
        registry.get_encoder("uint[]256")


def test_copying_copies_internal_mappings(registry: ABIRegistry):
    c1 = registry.copy()
    c2 = copy.copy(registry)
    c3 = copy.deepcopy(registry)

    for x, y in itertools.combinations((registry, c1, c2, c3), 2):
        # Objects are different
        assert id(x) != id(y)

        # Internal mapping objects are different
        assert id(x._encoders) != id(y._encoders)
        assert id(x._decoders) != id(y._decoders)

        # Coders can be found for existing registrations
        assert isinstance(x.get_encoder("address"), encoding.AddressEncoder)
        assert isinstance(x.get_decoder("address"), decoding.AddressDecoder)
        assert isinstance(y.get_encoder("address"), encoding.AddressEncoder)
        assert isinstance(y.get_decoder("address"), decoding.AddressDecoder)


def test_has_encoder_returns_true(registry: ABIRegistry):
    assert registry.has_encoder("address")


def test_has_encoder_raises(registry: ABIRegistry):
    registry.register(
        BaseEquals("address", with_sub=False),
        encoding.AddressEncoder,
        decoding.AddressDecoder,
        label="other_address",
    )

    with pytest.raises(exceptions.MultipleEntriesFound):
        assert registry.has_encoder("address")


def test_has_encoder_returns_false(registry: ABIRegistry):
    assert not registry.has_encoder("foo")



from hypothesis import (
    given,
    settings,
)

from faster_eth_abi import (
    decode,
    encode,
)
from tests.core.common.strategies import (
    multi_strs_values,
    single_strs_values,
    tuple_strs_values,
)


@settings(deadline=None)
@given(multi_strs_values)
def test_multi_abi_reversibility(types_and_values):
    """
    Tests round trip encoding and decoding for basic types and lists of basic
    types.
    """
    types, values = types_and_values
    encoded_values = encode(types, values)
    decoded_values = decode(types, encoded_values)
    assert values == decoded_values


@settings(deadline=None)
@given(single_strs_values)
def test_single_abi_type_reversibility(type_and_value):
    """
    Tests round trip encoding and decoding for basic types and lists of basic
    types.
    """
    _type, value = type_and_value
    encoded_value = encode([_type], [value])
    (decoded_value,) = decode([_type], encoded_value)
    assert value == decoded_value


@settings(deadline=None)
@given(tuple_strs_values)
def test_single_abi_type_tuple_reversibility(type_and_value):
    """
    Tests round trip encoding and decoding for tuple types.
    """
    _type, value = type_and_value
    encoded_value = encode([_type], [value])
    (decoded_value,) = decode([_type], encoded_value)
    assert value == decoded_value
